"""Thread pool module for running agents"""

from typing import TYPE_CHECKING, Tuple, Any, List
import threading
from queue import Queue
from enum import Enum

from langchain_core.runnables import Runnable

from motleycrew.common import Defaults

if TYPE_CHECKING:
    from motleycrew import Task
    from motleycrew.tasks import TaskUnit


class TaskUnitThreadState(Enum):
    BUSY = "busy"
    WAITING = "waiting"
    EXITED = "exited"


SENTINEL = object()  # sentinel object for closing threads


class TaskUnitThread(threading.Thread):
    def __init__(self, input_queue: Queue, output_queue: Queue, *args, **kwargs):
        """The thread class for running task units

        Args:
            input_queue (Queue): queue of task units to complete
            output_queue (Queue): queue of completed task units
            *args:
            **kwargs:
        """
        self.input_queue = input_queue
        self.output_queue = output_queue
        self._state = TaskUnitThreadState.WAITING

        super(TaskUnitThread, self).__init__(*args, **kwargs)

    @property
    def state(self):
        return self._state

    def run(self) -> None:
        while True:
            run_data = self.input_queue.get()
            self._state = TaskUnitThreadState.BUSY

            if run_data is SENTINEL:
                self._state = TaskUnitThreadState.EXITED
                self.input_queue.task_done()
                break

            agent, task, unit = run_data
            try:
                result = agent.invoke(unit.as_dict())
            except Exception as e:
                self.output_queue.put(e)
            else:
                self.output_queue.put((task, unit, result))
            finally:
                self._state = TaskUnitThreadState.WAITING
                self.input_queue.task_done()


class TaskUnitThreadPool:
    def __init__(self, num_threads: int = Defaults.DEFAULT_NUM_THREADS):
        """The thread pool class for performing task units

        Args:
            num_threads (int): number of threads to create
        """
        self.num_threads = num_threads

        self.input_queue = Queue()
        self.output_queue = Queue()

        self._threads = []
        for i in range(self.num_threads):
            thread = TaskUnitThread(self.input_queue, self.output_queue)
            thread.start()
            self._threads.append(thread)
        self._task_units_in_progress = []

    def add_task_unit(self, agent: Runnable, task: "Task", unit: "TaskUnit"):
        """Adds a task unit to the queue for execution

        Args:
            agent (Runnable):
            task (Task):
            unit (TaskUnit):

        Returns:

        """
        self._task_units_in_progress.append((task, unit))
        self.input_queue.put((agent, task, unit))

    def get_completed_task_units(self) -> List[Tuple["Task", "TaskUnit", Any]]:
        """Returns a list of completed task units with their results

        Returns:
            List[Tuple[Task, TaskUnit, Any]]: list of triplets of task, task unit, and result
        """
        completed_tasks = []
        while not self.output_queue.empty():
            task_result = self.output_queue.get()
            if isinstance(task_result, Exception):
                raise task_result

            task, unit, result = task_result
            completed_tasks.append((task, unit, result))
            self._task_units_in_progress.remove((task, unit))
        return completed_tasks

    def wait_and_close(self):
        """Wait for all task units to complete and close the threads"""
        for t in self._threads:
            if t.is_alive():
                self.input_queue.put(SENTINEL)
        self.input_queue.join()

        for t in self._threads:
            t.join()

    def is_completed(self) -> bool:
        """Returns whether all task units have been completed

        Returns:
            bool:
        """
        return not bool(self._task_units_in_progress)



""" Module description

Attributes:
    MotleyGraphNodeType (TypeVar):

"""

from typing import Optional, Any, TypeVar, TYPE_CHECKING
from pydantic import BaseModel

if TYPE_CHECKING:
    from motleycrew.storage import MotleyGraphStore


class MotleyGraphNode(BaseModel):
    """Description"""

    # Q: KuzuGraphNode a better name? Because def id is specific?
    # A: No, I think _id attribute is pretty universal
    __label__: Optional[str] = None
    __graph_store__: Optional["MotleyGraphStore"] = None

    @property
    def id(self) -> Optional[Any]:
        return getattr(self, "_id", None)

    @property
    def is_inserted(self) -> bool:
        return self.id is not None

    @classmethod
    def get_label(cls) -> str:
        """Description

        Returns:
            str:
        """
        # Q: why not @property def label(cls) -> str: return cls.__label__ or cls.__name__ ?
        # A: Because we want to be able to call this method without an instance
        #    and properties can't be class methods since Python 3.12
        if cls.__label__:
            return cls.__label__
        return cls.__name__

    def __setattr__(self, name, value):
        super().__setattr__(name, value)

        if name not in self.model_fields:
            # Q: Should we not raise an error here instead?
            # A: No, there are technical attributes like __graph_store__ that are not in the model
            return  # Non-pydantic field => not in the DB

        if self.__graph_store__ and self.is_inserted:
            self.__graph_store__.update_property(self, name)

    def __eq__(self, other):
        return self.is_inserted and self.get_label() == other.get_label() and self.id == other.id


MotleyGraphNodeType = TypeVar("MotleyGraphNodeType", bound=MotleyGraphNode)



""" Module description"""
from typing import Optional
from dataclasses import dataclass
import json

from motleycrew.storage.graph_node import MotleyGraphNode
from motleycrew.tasks import TaskUnit

REPR_CONTEXT_LENGTH_LIMIT = 30


class Question(MotleyGraphNode):
    """ Description

    Attributes:
        question (str):
        answer (:obj:`str`, optional):
        context (:obj:`List[str]`, optional)

    """
    question: str
    answer: Optional[str] = None
    context: Optional[list[str]] = None

    def __repr__(self):
        if self.context and len(self.context):
            context_repr = '", "'.join(self.context)
            if len(context_repr) > REPR_CONTEXT_LENGTH_LIMIT:
                context_repr = '["' + context_repr[:REPR_CONTEXT_LENGTH_LIMIT] + "...]"
            else:
                context_repr = '["' + context_repr + '"]'
        else:
            context_repr = str(self.context)

        return "Question(id={}, question={}, answer={}, context={})".format(
            self.id, self.question, self.answer, context_repr
        )


class QuestionGenerationTaskUnit(TaskUnit):
    question: Question


class QuestionAnsweringTaskUnit(TaskUnit):
    question: Question



from typing import Collection, Sequence, Optional, Any, List, Tuple
import os
import asyncio
import threading
import time

from motleycrew.agents.parent import MotleyAgentParent
from motleycrew.tasks import Task, TaskUnit, SimpleTask, TaskUnitType
from motleycrew.storage import MotleyGraphStore
from motleycrew.storage.graph_store_utils import init_graph_store
from motleycrew.tools import MotleyTool
from motleycrew.common import logger, AsyncBackend, Defaults
from motleycrew.crew.crew_threads import TaskUnitThreadPool


class MotleyCrew:
    _loop: Optional[asyncio.AbstractEventLoop] = None

    def __init__(
        self,
        graph_store: Optional[MotleyGraphStore] = None,
        async_backend: AsyncBackend = AsyncBackend.NONE,
        num_threads: int = Defaults.DEFAULT_NUM_THREADS,
    ):
        if graph_store is None:
            graph_store = init_graph_store()
        self.graph_store = graph_store

        self.tools = []
        self.tasks = []
        self.async_backend = async_backend
        self.num_threads = num_threads

    def create_simple_task(
        self,
        description: str,
        agent: MotleyAgentParent,
        name: Optional[str] = None,
        generate_name: bool = False,
        tools: Optional[Sequence[MotleyTool]] = None,
    ) -> SimpleTask:
        """
        Basic method for creating a simple task
        """
        if name is None and generate_name:
            # Call llm to generate a name
            raise NotImplementedError("Name generation not yet implemented")
        task = SimpleTask(crew=self, name=name, description=description, agent=agent, tools=tools)
        self.register_tasks([task])
        return task

    def run(self) -> list[TaskUnit]:
        if self.async_backend == AsyncBackend.NONE:
            result = self._run_sync()
        elif self.async_backend == AsyncBackend.ASYNCIO:
            try:
                result = asyncio.run(self._run_async())
            except RuntimeError:
                if self._loop is None:
                    self._loop = asyncio.new_event_loop()
                    threading.Thread(target=self._loop.run_forever, daemon=True).start()

                future: asyncio.Future = asyncio.run_coroutine_threadsafe(
                    self._run_async(), self._loop
                )
                result = future.result()
        elif self.async_backend == AsyncBackend.THREADING:
            result = self._run_threaded()
        else:
            raise NotImplementedError()

        return result

    def add_dependency(self, upstream: Task, downstream: Task):
        self.graph_store.create_relation(
            upstream.node, downstream.node, label=Task.TASK_IS_UPSTREAM_LABEL
        )
        # # TODO: rollback if bad?
        # self.check_cyclical_dependencies()

    def register_tasks(self, tasks: Collection[Task]):
        for task in tasks:
            if task not in self.tasks:
                self.tasks.append(task)
                task.crew = self
                task.prepare_graph_store()
                self.graph_store.insert_node(task.node)

                self.graph_store.ensure_relation_table(
                    from_class=type(task.node),
                    to_class=type(task.node),
                    label=Task.TASK_IS_UPSTREAM_LABEL,
                )  # TODO: remove this workaround, https://github.com/kuzudb/kuzu/issues/3488

    def _prepare_next_unit_for_dispatch(
        self, running_sync_tasks: set
    ) -> Tuple[MotleyAgentParent, Task, TaskUnit]:
        """Retrieve and prepare the next unit for dispatch
        Args:
            running_sync_tasks (set): Collection of currently running forced synchronous tasks

        Yields:
            tuple: agent, task, unit to be dispatched
        """
        available_tasks = self.get_available_tasks()
        logger.info("Available tasks: %s", available_tasks)

        for task in available_tasks:
            if not task.allow_async_units and task in running_sync_tasks:
                continue

            logger.info("Processing task: %s", task)

            next_unit = task.get_next_unit()

            if next_unit is None:
                logger.info("Got no matching units for task %s", task)
                continue

            if not task.allow_async_units:
                running_sync_tasks.add(task)

            logger.info("Got a matching unit for task %s", task)
            logger.info("Processing unit: %s", next_unit)

            extra_tools = self.get_extra_tools(task)
            agent = task.get_worker(extra_tools)

            logger.info("Assigned unit %s to agent %s, dispatching", next_unit, agent)
            next_unit.set_running()
            self.add_task_unit_to_graph(task=task, unit=next_unit)
            task.register_started_unit(next_unit)

            yield agent, task, next_unit

    def _handle_task_unit_completion(
        self,
        task: Task,
        unit: TaskUnit,
        result: Any,
        running_sync_tasks: set,
        done_units: list[TaskUnit],
    ):
        """Handle task unit completion
        Args:
            task (Task): Task object
            unit (TaskUnit): Task unit object
            result (Any): Result of the task unit
            running_sync_tasks (set): Collection of currently running forced synchronous tasks
            done_units (list): List of completed task units

        """
        if task in running_sync_tasks:
            running_sync_tasks.remove(task)

        unit.output = result
        logger.info("Task unit %s completed, marking as done", unit)
        unit.set_done()
        task.register_completed_unit(unit)
        done_units.append(unit)

    def _run_sync(self) -> list[TaskUnit]:
        """Run crew synchronously

        Returns:
            :obj:`list` of :obj:`TaskUnit`:
        """
        done_units = []
        while True:
            did_something = False

            available_tasks = self.get_available_tasks()
            logger.info("Available tasks: %s", available_tasks)

            for agent, task, unit in self._prepare_next_unit_for_dispatch(set()):
                # TODO: accept and handle some sort of return value? Or just the final state of the task?
                result = agent.invoke(unit.as_dict())

                self._handle_task_unit_completion(
                    task=task,
                    unit=unit,
                    result=result,
                    running_sync_tasks=set(),
                    done_units=done_units,
                )

                did_something = True

            if not did_something:
                logger.info("Nothing left to do, exiting")
                return done_units

    def _run_threaded(self) -> list[TaskUnit]:
        """Run crew in threads

        Returns:
            :obj:`list` of :obj:`TaskUnit`:
        """

        done_units = []
        running_sync_tasks = set()
        thread_pool = TaskUnitThreadPool(self.num_threads)
        try:
            while True:
                for task, unit, result in thread_pool.get_completed_task_units():
                    self._handle_task_unit_completion(
                        task=task,
                        unit=unit,
                        result=result,
                        running_sync_tasks=running_sync_tasks,
                        done_units=done_units,
                    )

                for agent, next_task, next_unit in self._prepare_next_unit_for_dispatch(
                    running_sync_tasks
                ):
                    thread_pool.add_task_unit(agent, next_task, next_unit)

                if thread_pool.is_completed():
                    logger.info("Nothing left to do, exiting")
                    return done_units

                time.sleep(Defaults.DEFAULT_EVENT_LOOP_SLEEP)
        finally:
            thread_pool.wait_and_close()

    @staticmethod
    async def _async_invoke_agent(agent: MotleyAgentParent, unit: TaskUnit) -> Any:
        return await agent.ainvoke(unit.as_dict())

    async def _run_async(self) -> list[TaskUnit]:
        """Run crew asynchronously

        Returns:
            :obj:`list` of :obj:`TaskUnit`:
        """

        done_units = []
        async_units = {}
        running_tasks = set()

        while True:
            for async_task in list(async_units.keys()):
                if async_task.done():
                    task, unit = async_units.pop(async_task)

                    self._handle_task_unit_completion(
                        task=task,
                        unit=unit,
                        result=async_task.result(),
                        running_sync_tasks=running_tasks,
                        done_units=done_units,
                    )

            for agent, next_task, next_unit in self._prepare_next_unit_for_dispatch(running_tasks):
                async_task = asyncio.create_task(MotleyCrew._async_invoke_agent(agent, next_unit))
                async_units[async_task] = (next_task, next_unit)

            if not async_units:
                logger.info("Nothing left to do, exiting")
                return done_units

            await asyncio.sleep(Defaults.DEFAULT_EVENT_LOOP_SLEEP)

    def get_available_tasks(self) -> list[Task]:
        query = (
            "MATCH (downstream:{}) "
            "WHERE NOT downstream.done "
            "AND NOT EXISTS {{MATCH (upstream:{})-[:{}]->(downstream) "
            "WHERE NOT upstream.done}} "
            "RETURN downstream"
        ).format(
            Task.NODE_CLASS.get_label(),
            Task.NODE_CLASS.get_label(),
            Task.TASK_IS_UPSTREAM_LABEL,
        )
        available_task_nodes = self.graph_store.run_cypher_query(query, container=Task.NODE_CLASS)
        return [task for task in self.tasks if task.node in available_task_nodes]

    def add_task_unit_to_graph(self, task: Task, unit: TaskUnitType) -> None:
        """Add a task unit to the graph and connect it to its task

        Args:
            task (Task):
            unit (TaskUnitType):

        Returns:

        """
        assert isinstance(unit, task.task_unit_class)
        assert not unit.done

        self.graph_store.upsert_triplet(
            from_node=unit,
            to_node=task.node,
            label=task.task_unit_belongs_label,
        )

    def get_extra_tools(self, task: Task) -> list[MotleyTool]:
        # TODO: Smart tool selection goes here
        tools = []
        tools += self.tools or []
        # tools += task.tools or []

        return tools

    def check_cyclical_dependencies(self):
        pass  # TODO



""" Module description"""
from typing import Optional, Type

from langchain_core.tools import StructuredTool
from langchain_core.prompts import PromptTemplate
from langchain_core.prompts.base import BasePromptTemplate
from langchain_core.language_models import BaseLanguageModel
from langchain_core.pydantic_v1 import BaseModel, Field, create_model

from motleycrew.tools import MotleyTool
from motleycrew.common import LLMFramework
from motleycrew.common.llms import init_llm


class LLMTool(MotleyTool):
    def __init__(
        self,
        name: str,
        description: str,
        prompt: str | BasePromptTemplate,
        llm: Optional[BaseLanguageModel] = None,
        input_schema: Optional[Type[BaseModel]] = None,
    ):
        """ Description

        Args:
            name (str):
            description (str):
            prompt (:obj:`str`, :obj:`BasePromptTemplate`):
            llm (:obj:`BaseLanguageModel`, optional):
            input_schema (:obj:`Type[BaseModel]`, optional):
        """
        langchain_tool = create_llm_langchain_tool(
            name=name,
            description=description,
            prompt=prompt,
            llm=llm,
            input_schema=input_schema,
        )
        super().__init__(langchain_tool)


def create_llm_langchain_tool(
    name: str,
    description: str,
    prompt: str | BasePromptTemplate,
    llm: Optional[BaseLanguageModel] = None,
    input_schema: Optional[Type[BaseModel]] = None,
):
    """ Description

    Args:
        name (str):
        description (str):
        prompt (:obj:`str`, :obj:`BasePromptTemplate`):
        llm (:obj:`BaseLanguageModel`, optional):
        input_schema (:obj:`Type[BaseModel]`, optional):

    Returns:

    """
    if llm is None:
        llm = init_llm(llm_framework=LLMFramework.LANGCHAIN)

    if not isinstance(prompt, BasePromptTemplate):
        prompt = PromptTemplate.from_template(prompt)

    if input_schema is None:
        fields = {
            var: (str, Field(description=f"Input {var} for the tool."))
            for var in prompt.input_variables
        }

        # Create the LLMToolInput class dynamically
        input_schema = create_model("LLMToolInput", **fields)

    def call_llm(**kwargs) -> str:
        chain = prompt | llm
        return chain.invoke(kwargs)

    return StructuredTool.from_function(
        func=call_llm,
        name=name,
        description=description,
        args_schema=input_schema,
    )



""" Module description

Attributes:
    prompt_template (str):
    dall_e_template (str):
"""
from typing import Optional

import os
import requests
import mimetypes

from langchain.agents import Tool
from langchain_core.pydantic_v1 import BaseModel, Field
from langchain_community.utilities.dalle_image_generator import DallEAPIWrapper

from motleycrew.tools.tool import MotleyTool
import motleycrew.common.utils as motley_utils
from motleycrew.common import LLMFramework
from motleycrew.common.llms import init_llm
from motleycrew.common import logger
from langchain.prompts import PromptTemplate


def download_image(url: str, file_path: str) -> Optional[str]:
    """ Description

    Args:
        url (str):
        file_path (str):

    Returns:
        :obj:`str`, None:
    """
    response = requests.get(url, stream=True)
    if response.status_code == requests.codes.ok:
        content_type = response.headers.get("content-type")
        extension = mimetypes.guess_extension(content_type)
        if not extension:
            extension = ".png"  # default to .png if content-type is not recognized

        file_path_with_extension = file_path + extension
        logger.info("Downloading image %s to %s", url, file_path_with_extension)

        with open(file_path_with_extension, "wb") as f:
            for chunk in response:
                f.write(chunk)

        return file_path_with_extension
    else:
        logger.error("Failed to download image. Status code: %s", response.status_code)


class DallEImageGeneratorTool(MotleyTool):
    def __init__(
        self,
        images_directory: Optional[str] = None,
        refine_prompt_with_llm: bool = True,
        model: str = "dall-e-3",
        quality: str = "standard",
        size: str = "1024x1024",
        style: Optional[str] = None,
    ):
        """ Description

        Args:
            images_directory (:obj:`str`, optional):
            refine_prompt_with_llm (:obj:`bool`, optional):
            model (:obj:`str`, optional):
            quality (:obj:`str`, optional):
            size (:obj:`str`, optional):
            style (:obj:`str`, optional):
        """
        langchain_tool = create_dalle_image_generator_langchain_tool(
            images_directory=images_directory,
            refine_prompt_with_llm=refine_prompt_with_llm,
            model=model,
            quality=quality,
            size=size,
            style=style,
        )
        super().__init__(langchain_tool)


class DallEToolInput(BaseModel):
    """Input for the Dall-E tool.

    Attributes:
        description (str):
    """

    description: str = Field(description="image description")


prompt_template = """Generate a detailed DALL-E prompt to generate an image 
based on the following description: 
```{text}```
Your output MUST NOT exceed 3500 characters"""

dall_e_template = """{text}
Note: Do not include any text in the images.
"""


def run_dalle_and_save_images(
    description: str,
    images_directory: Optional[str] = None,
    refine_prompt_with_llm: bool = True,
    model: str = "dall-e-3",
    quality: str = "standard",
    size: str = "1024x1024",
    style: Optional[str] = None,
    file_name_length: int = 8,
) -> Optional[list[str]]:
    """ Description

    Args:
        description (str):
        images_directory (:obj:`str`, optional):
        refine_prompt_with_llm(:obj:`bool`, optional):
        model (:obj:`str`, optional):
        quality (:obj:`str`, optional):
        size (:obj:`str`, optional):
        style (:obj:`str`, optional):
        file_name_length (:obj:`int`, optional):

    Returns:
        :obj:`list` of :obj:`str`:
    """

    dall_e_prompt = PromptTemplate.from_template(dall_e_template)

    if refine_prompt_with_llm:
        prompt = PromptTemplate.from_template(template=prompt_template)
        llm = init_llm(llm_framework=LLMFramework.LANGCHAIN)
        dall_e_prompt = prompt | llm | (lambda x: {"text": x.content}) | dall_e_prompt

    prompt_value = dall_e_prompt.invoke({"text": description})

    dalle_api = DallEAPIWrapper(
        model=model,
        quality=quality,
        size=size,
        model_kwargs={"style": style} if (model == "dall-e-3" and style) else {},
    )

    dalle_result = dalle_api.run(prompt_value.text)
    logger.info("Dall-E API output: %s", dalle_result)

    urls = dalle_result.split(dalle_api.separator)
    if not len(urls) or not motley_utils.is_http_url(urls[0]):
        logger.error("Dall-E API did not return a valid url: %s", dalle_result)
        return

    if images_directory:
        os.makedirs(images_directory, exist_ok=True)
        file_paths = []
        for url in urls:
            file_name = motley_utils.generate_hex_hash(url, length=file_name_length)
            file_path = os.path.join(images_directory, file_name)

            file_path_with_extension = download_image(url=url, file_path=file_path).replace(
                os.sep, "/"
            )
            file_paths.append(file_path_with_extension)
        return file_paths
    else:
        logger.info("Images directory is not provided, returning URLs")
        return urls


def create_dalle_image_generator_langchain_tool(
    images_directory: Optional[str] = None,
    refine_prompt_with_llm: bool = True,
    model: str = "dall-e-3",
    quality: str = "standard",
    size: str = "1024x1024",
    style: Optional[str] = None,
):
    """ Description

    Args:
        images_directory (:obj:`str`, optional):
        refine_prompt_with_llm (:obj:`bool`, optional):
        model (:obj:`str`, optional):
        quality (:obj:`str`, optional):
        size (:obj:`str`, optional):
        style (:obj:`str`, optional):

    Returns:
        Tool:
    """
    def run_dalle_and_save_images_partial(description: str):
        return run_dalle_and_save_images(
            description=description,
            images_directory=images_directory,
            refine_prompt_with_llm=refine_prompt_with_llm,
            model=model,
            quality=quality,
            size=size,
            style=style,
        )

    return Tool(
        name="dalle_image_generator",
        func=run_dalle_and_save_images_partial,
        description="A wrapper around OpenAI DALL-E API. Useful for when you need to generate images from a text description. "
        "Input should be an image description.",
        args_schema=DallEToolInput,
    )


if __name__ == "__main__":
    tool = DallEImageGeneratorTool()
    out = tool.invoke("A beautiful castle on top of a hill at sunset")
    logger.info(out)



""" Module description """

from typing import Sequence

try:
    from llama_index.core.agent import ReActAgent
    from llama_index.core.llms import LLM
    from llama_index.core.callbacks import CallbackManager
except ImportError:
    LLM = object

from motleycrew.agents.llama_index import LlamaIndexMotleyAgent
from motleycrew.tools import MotleyTool
from motleycrew.common import MotleySupportedTool
from motleycrew.common import LLMFramework
from motleycrew.common.llms import init_llm
from motleycrew.tracking import get_default_callbacks_list
from motleycrew.common.utils import ensure_module_is_installed


class ReActLlamaIndexMotleyAgent(LlamaIndexMotleyAgent):
    def __init__(
        self,
        description: str | None = None,
        name: str | None = None,
        tools: Sequence[MotleySupportedTool] | None = None,
        llm: LLM | None = None,
        verbose: bool = False,
    ):
        """Description

        Args:
            description (:obj:`str`, optional):
            name (:obj:`str`, optional):
            tools (:obj:`Sequence[MotleySupportedTool]`, optional):
            llm (:obj:`LLM`, optional):
            verbose (:obj:`bool`, optional):
        """
        ensure_module_is_installed("llama_index")
        if llm is None:
            llm = init_llm(llm_framework=LLMFramework.LLAMA_INDEX)

        def agent_factory(tools: dict[str, MotleyTool]) -> ReActAgent:
            llama_index_tools = [t.to_llama_index_tool() for t in tools.values()]
            callbacks = get_default_callbacks_list(LLMFramework.LLAMA_INDEX)
            agent = ReActAgent.from_tools(
                tools=llama_index_tools,
                llm=llm,
                verbose=verbose,
                callback_manager=CallbackManager(callbacks),
            )
            return agent

        super().__init__(
            description=description,
            name=name,
            agent_factory=agent_factory,
            tools=tools,
            verbose=verbose,
        )



from motleycrew.tasks.task_unit import TaskUnit
from motleycrew.tasks.task_unit import TaskUnitType
from motleycrew.tasks.task import Task
from motleycrew.tasks.simple import SimpleTask

__all__ = ["TaskUnit", "TaskUnitType", "Task", "SimpleTask"]



""" Module description """

from typing import Union, Optional, Dict, Any
from typing import Callable

from langchain.tools import BaseTool
from langchain_core.runnables import Runnable, RunnableConfig

try:
    from llama_index.core.tools import BaseTool as LlamaIndex__BaseTool
    from llama_index.core.tools import FunctionTool as LlamaIndex__FunctionTool
except ImportError:
    LlamaIndex__BaseTool = None
    LlamaIndex__FunctionTool = None

from motleycrew.common.utils import ensure_module_is_installed
from motleycrew.common.types import MotleySupportedTool
from motleycrew.agents.abstract_parent import MotleyAgentAbstractParent


class MotleyTool(Runnable):

    def __init__(self, tool: BaseTool):
        """Base tool class compatible with MotleyAgents.
        It is a wrapper for LangChain's BaseTool, containing all necessary adapters and converters.

        Args:
            tool (BaseTool):
        """
        self.tool = tool

    def __repr__(self):
        return f"MotleyTool(name={self.name})"

    def __str__(self):
        return self.__repr__()

    @property
    def name(self):
        # TODO: do we really want to make a thin wrapper in this fashion?
        return self.tool.name

    @property
    def description(self):
        return self.tool.description

    @property
    def args_schema(self):
        return self.tool.args_schema

    def invoke(
        self,
        input: Union[str, Dict],
        config: Optional[RunnableConfig] = None,
        **kwargs: Any,
    ) -> Any:
        return self.tool.invoke(input=input, config=config, **kwargs)

    @staticmethod
    def from_langchain_tool(langchain_tool: BaseTool) -> "MotleyTool":
        """Description

        Args:
            langchain_tool (BaseTool):

        Returns:
            MotleyTool:
        """
        return MotleyTool(tool=langchain_tool)

    @staticmethod
    def from_llama_index_tool(llama_index_tool: LlamaIndex__BaseTool) -> "MotleyTool":
        """Description

        Args:
            llama_index_tool (LlamaIndex__BaseTool):

        Returns:
            MotleyTool:
        """
        ensure_module_is_installed("llama_index")
        langchain_tool = llama_index_tool.to_langchain_tool()
        return MotleyTool.from_langchain_tool(langchain_tool=langchain_tool)

    @staticmethod
    def from_supported_tool(tool: MotleySupportedTool) -> "MotleyTool":
        """Description

        Args:
            tool (:obj:`MotleyTool`, :obj:`BaseTool`, :obj:`LlamaIndex__BaseTool`, :obj:`MotleyAgentAbstractParent`):

        Returns:

        """
        if isinstance(tool, MotleyTool):
            return tool
        elif isinstance(tool, BaseTool):
            return MotleyTool.from_langchain_tool(tool)
        elif isinstance(tool, LlamaIndex__BaseTool):
            return MotleyTool.from_llama_index_tool(tool)
        elif isinstance(tool, MotleyAgentAbstractParent):
            return tool.as_tool()
        else:
            raise Exception(
                f"Tool type `{type(tool)}` is not supported, please convert to MotleyTool first"
            )

    def to_langchain_tool(self) -> BaseTool:
        """Description

        Returns:
            BaseTool:
        """
        return self.tool

    def to_llama_index_tool(self) -> LlamaIndex__BaseTool:
        """Description

        Returns:
            LlamaIndex__BaseTool:
        """
        ensure_module_is_installed("llama_index")
        llama_index_tool = LlamaIndex__FunctionTool.from_defaults(
            fn=self.tool._run,
            name=self.tool.name,
            description=self.tool.description,
            fn_schema=self.tool.args_schema,
        )
        return llama_index_tool

    def to_autogen_tool(self) -> Callable:
        """Description

        Returns:
            Callable:
        """
        fields = list(self.tool.args_schema.__fields__.values())
        if len(fields) != 1:
            raise Exception("Multiple input fields are not supported in to_autogen_tool")

        field_name = fields[0].name
        field_type = fields[0].annotation

        def autogen_tool_fn(input: field_type) -> str:
            return self.invoke({field_name: input})

        return autogen_tool_fn



""" Module description

Attributes:
    _default_prompt (PromptTemplate):
"""
from langchain_core.pydantic_v1 import BaseModel, Field
from langchain.prompts import PromptTemplate
from langchain_core.prompts.base import BasePromptTemplate
from langchain_core.tools import StructuredTool
from langchain_core.runnables import (
    RunnablePassthrough,
    RunnableLambda,
    chain,
)

from motleycrew.tools import MotleyTool
from motleycrew.tools import LLMTool
from motleycrew.common.utils import print_passthrough

from motleycrew.applications.research_agent.question import Question


class QuestionPrioritizerTool(MotleyTool):
    def __init__(
        self,
        prompt: str | BasePromptTemplate = None,
    ):
        """ Description

        Args:
            prompt (:obj:`str`, :obj:`BasePromptTemplate`, optional):
        """
        langchain_tool = create_question_prioritizer_langchain_tool(prompt=prompt)

        super().__init__(langchain_tool)


_default_prompt = PromptTemplate(
    template=(
        "You are provided with the following list of questions:"
        " {unanswered_questions_text} \n"
        " Your task is to choose one question from the above list"
        " that is the most pertinent to the following query:\n"
        " '{original_question_text}' \n"
        " Respond with a single number the chosen question out of the provided list of questions."
        " Return only the number as it is without any edits."
    ),
    input_variables=["unanswered_questions", "original_question"],
)


class QuestionPrioritizerInput(BaseModel, arbitrary_types_allowed=True):
    """ Description

    Attributes:
        original_question (Question):
        unanswered_questions (list[Question]):

    """
    original_question: Question = Field(description="The original question.")
    unanswered_questions: list[Question] = Field(
        description="Questions to pick the most pertinent to the original question from.",
    )


def create_question_prioritizer_langchain_tool(
    prompt: str | BasePromptTemplate = None,
) -> StructuredTool:
    """ Creates a LangChainTool for the AnswerSubQuestionTool.

    Args:
        prompt (:obj:`str`, :obj:`BasePromptTemplate`, optional):

    Returns:
        StructuredTool
    """
    if prompt is None:
        prompt = _default_prompt

    question_prioritizer = LLMTool(
        prompt=prompt,
        name="Question prioritizer",
        description="Takes the original question and a list of derived questions, "
        "and selects from the latter the one most pertinent to the former",
    )

    @chain
    def get_original_question_text(input_dict: dict):
        return input_dict["original_question"].question

    @chain
    def format_unanswered_questions(input_dict: dict):
        unanswered_questions: list[Question] = input_dict["unanswered_questions"]
        return "\n".join(
            "{}. {}".format(i + 1, question.question)
            for i, question in enumerate(unanswered_questions)
        )

    @chain
    def get_most_pertinent_question(input_dict: dict):
        unanswered_questions: list[Question] = input_dict["unanswered_questions"]
        most_pertinent_question_id = (
            int(input_dict["most_pertinent_question_id_message"].content.strip(" \n.")) - 1
        )
        assert most_pertinent_question_id < len(unanswered_questions)
        return unanswered_questions[most_pertinent_question_id]

    this_chain = (
        RunnablePassthrough.assign(
            original_question_text=lambda x: x["original_question"].question,
            unanswered_questions_text=format_unanswered_questions,
        )
        | RunnableLambda(print_passthrough)
        | RunnablePassthrough.assign(
            most_pertinent_question_id_message=question_prioritizer.to_langchain_tool()
        )
        | RunnableLambda(print_passthrough)
        | get_most_pertinent_question
    )

    langchain_tool = StructuredTool.from_function(
        func=lambda original_question, unanswered_questions: this_chain.invoke(
            {"original_question": original_question, "unanswered_questions": unanswered_questions}
        ),
        name=question_prioritizer.name,
        description=question_prioritizer.tool.description,
        args_schema=QuestionPrioritizerInput,
    )

    return langchain_tool


if __name__ == "__main__":
    q = Question(question="What color is the sky?")
    unanswered = [
        Question(question="What time of day is it?"),
        Question(question="Who was H.P.Lovecraft?"),
    ]

    out = QuestionPrioritizerTool().invoke(
        {"unanswered_questions": unanswered, "original_question": q}
    )
    print(out)
    print("yay!")



""" Module description """
from motleycrew.common import LLMFamily
from motleycrew.common import GraphStoreType


class Defaults:
    """ Description

    Attributes:
        DEFAULT_LLM_FAMILY (str):
        DEFAULT_LLM_NAME (str):
        DEFAULT_LLM_TEMPERATURE (float):
        LLM_MAP (dict):
        DEFAULT_GRAPH_STORE_TYPE (str):
        MODULE_INSTALL_COMMANDS (dict):
        DEFAULT_NUM_THREADS (int):
        DEFAULT_EVENT_LOOP_SLEEP (int):

    """
    DEFAULT_LLM_FAMILY = LLMFamily.OPENAI
    DEFAULT_LLM_NAME = "gpt-4o"
    DEFAULT_LLM_TEMPERATURE = 0.0
    LLM_MAP = {}

    DEFAULT_GRAPH_STORE_TYPE = GraphStoreType.KUZU

    MODULE_INSTALL_COMMANDS = {
        "crewai": "pip install crewai",
        "llama_index": "pip install llama-index",
        "autogen": "pip install pyautogen",
        "lunary": "pip install lunary",
        "aider": "pip install aider-chat",
        "pglast": "pip install pglast",
    }

    DEFAULT_NUM_THREADS = 4
    DEFAULT_EVENT_LOOP_SLEEP = 1



""" Module description """
from langchain.agents import Tool
from langchain_experimental.utilities import PythonREPL
from langchain_core.pydantic_v1 import BaseModel, Field

from .tool import MotleyTool


class PythonREPLTool(MotleyTool):
    def __init__(self):
        """ Description

        """
        langchain_tool = create_repl_tool()
        super().__init__(langchain_tool)


class REPLToolInput(BaseModel):
    """Input for the REPL tool.

    Attributes:
        command (str):
    """

    command: str = Field(description="code to execute")


# You can create the tool to pass to an agent
def create_repl_tool():
    """ Description

    Returns:
        Tool:
    """
    return Tool.from_function(
        func=PythonREPL().run,
        name="python_repl",
        description="A Python shell. Use this to execute python commands. Input should be a valid python command. "
        "MAKE SURE TO PRINT OUT THE RESULTS YOU CARE ABOUT USING `print(...)`.",
        args_schema=REPLToolInput,
    )



from langchain_core.tools import Tool
from langchain_core.pydantic_v1 import BaseModel, Field

try:
    from pglast import parse_sql, prettify
    from pglast.parser import ParseError
except ImportError:
    parse_sql = None
    prettify = None
    ParseError = None

from motleycrew.tools import MotleyTool
from motleycrew.common.utils import ensure_module_is_installed


class PostgreSQLLinterTool(MotleyTool):

    def __init__(self):
        """PostgreSQL code verification tool
        """
        ensure_module_is_installed("pglast")

        langchain_tool = create_pgsql_linter_tool()
        super().__init__(langchain_tool)


class PostgreSQLLinterInput(BaseModel):
    """Input for the PostgreSQLLinterTool.

    Attributes:
        query (str):
    """

    query: str = Field(description="SQL code for validation")


def create_pgsql_linter_tool() -> Tool:
    """Create the underlying langchain tool for PostgreSQLLinterTool

    Returns:
        Tool:
    """
    def parse_func(query: str) -> str:
        try:
            parse_sql(query)
            return prettify(query)
        except ParseError as e:
            return str(e)

    return Tool.from_function(
        func=parse_func,
        name="PostgreSQL linter tool",
        description="Tool for validating PostgreSQL code",
        args_schema=PostgreSQLLinterInput,
    )



""" Module description """

from typing import TYPE_CHECKING, Optional, Sequence, Any, Callable, Dict, Union, Tuple
from functools import wraps
import inspect

from langchain_core.tools import Tool
from langchain_core.runnables import Runnable
from langchain_core.prompts.chat import ChatPromptTemplate, HumanMessage, SystemMessage
from langchain_core.tools import StructuredTool
from pydantic import BaseModel

from motleycrew.agents.abstract_parent import MotleyAgentAbstractParent
from motleycrew.tools import MotleyTool
from motleycrew.common import MotleyAgentFactory, MotleySupportedTool
from motleycrew.common.exceptions import (
    AgentNotMaterialized,
    CannotModifyMaterializedAgent,
    InvalidOutput,
)
from motleycrew.common import logger

if TYPE_CHECKING:
    from motleycrew import MotleyCrew


class DirectOutput(BaseException):
    def __init__(self, output: Any):
        self.output = output


class MotleyAgentParent(MotleyAgentAbstractParent, Runnable):
    def __init__(
        self,
        description: str | None = None,
        name: str | None = None,
        agent_factory: MotleyAgentFactory | None = None,
        tools: Sequence[MotleySupportedTool] | None = None,
        output_handler: MotleySupportedTool | None = None,
        verbose: bool = False,
    ):
        """Description

        Args:
            description (:obj:`str`, optional):
            name (:obj:`str`, optional):
            agent_factory (:obj:`MotleyAgentFactory`, optional):
            tools (:obj:`Sequence[MotleySupportedTool]`, optional):
            output_handler (:obj:`MotleySupportedTool`, optional):
            verbose (:obj:`bool`, optional):
        """
        self.name = name or description
        self.description = description  # becomes tool description
        self.agent_factory = agent_factory
        self.tools: dict[str, MotleyTool] = {}
        self.output_handler = output_handler
        self.verbose = verbose
        self.crew: MotleyCrew | None = None

        self._agent = None

        if tools:
            self.add_tools(tools)

    def __repr__(self):
        return f"Agent(name={self.name})"

    def __str__(self):
        return self.__repr__()

    def compose_prompt(self, input_dict: dict, prompt: ChatPromptTemplate | str) -> str:
        # TODO: always cast description and prompt to ChatPromptTemplate first?
        prompt_messages = []

        if not self.description and not prompt:
            raise Exception("Cannot compose agent prompt without description or prompt")

        if self.description:
            if isinstance(self.description, ChatPromptTemplate):
                prompt_messages += self.description.invoke(input_dict).to_messages()

            elif isinstance(self.description, str):
                prompt_messages.append(SystemMessage(content=self.description))

            else:
                raise ValueError("Agent description must be a string or a ChatPromptTemplate")

        if prompt:
            if isinstance(prompt, ChatPromptTemplate):
                prompt_messages += prompt.invoke(input_dict).to_messages()

            elif isinstance(prompt, str):
                prompt_messages.append(HumanMessage(content=prompt))

            else:
                raise ValueError("Prompt must be a string or a ChatPromptTemplate")

        prompt = "\n\n".join([m.content for m in prompt_messages]) + "\n"
        return prompt

    @property
    def agent(self):
        """
        Getter for the inner agent that makes sure it's already materialized.
        The inner agent should always be accessed via this property method.
        """
        if not self.is_materialized:
            raise AgentNotMaterialized(agent_name=self.name)
        return self._agent

    @property
    def is_materialized(self):
        return self._agent is not None

    def _prepare_output_handler(self) -> Optional[MotleyTool]:
        """
        Wraps the output handler in one more tool layer,
        adding the necessary stuff for returning direct output through output handler.
        Expects agent's later invocation through _run_and_catch_output method below.
        """
        if not self.output_handler:
            return None

        def handle_agent_output(*args, **kwargs):
            assert self.output_handler
            try:
                output = self.output_handler._run(*args, **kwargs)
            except InvalidOutput as exc:
                return f"{exc.__class__.__name__}: {str(exc)}"

            raise DirectOutput(output)

        description = self.output_handler.description or f"Output handler for {self.name}"
        assert isinstance(description, str)
        description += "\n ONLY RETURN THE FINAL RESULT USING THIS TOOL!"

        prepared_output_handler = StructuredTool.from_function(
            name=self.output_handler.name,
            description=description,
            func=handle_agent_output,
            args_schema=self.output_handler.args_schema,
        )

        return MotleyTool.from_langchain_tool(prepared_output_handler)

    @staticmethod
    def _run_and_catch_output(
        action: Callable, action_args: tuple, action_kwargs: Dict[str, Any]
    ) -> Tuple[bool, Any]:
        """
        Catcher for the direct output from the output handler (see _prepare_output_handler).

        Args:
            action (Callable): the action inside which the output handler is supposed to be called.
                Usually the invocation method of the underlying agent.
            action_args (tuple): the args for the action
            action_kwargs (tuple): the kwargs for the action

        Returns:
            tuple[bool, Any]: a tuple with a boolean indicating whether the output was caught
                via DirectOutput and the output itself
        """
        assert callable(action)

        try:
            output = action(*action_args, **action_kwargs)
        except DirectOutput as output_exc:
            return True, output_exc.output

        return False, output

    def materialize(self):
        if self.is_materialized:
            logger.info("Agent is already materialized, skipping materialization")
            return
        assert self.agent_factory, "Cannot materialize agent without a factory provided"

        output_handler = self._prepare_output_handler()

        if inspect.signature(self.agent_factory).parameters.get("output_handler"):
            logger.info("Agent factory accepts output handler, passing it")
            self._agent = self.agent_factory(tools=self.tools, output_handler=output_handler)
        elif output_handler:
            logger.info("Agent factory does not accept output handler, passing it as a tool")
            tools_with_output_handler = self.tools.copy()
            tools_with_output_handler[output_handler.name] = output_handler
            self._agent = self.agent_factory(tools=tools_with_output_handler)
        else:
            self._agent = self.agent_factory(tools=self.tools)

    def add_tools(self, tools: Sequence[MotleySupportedTool]):
        """Description

        Args:
            tools (Sequence[MotleySupportedTool]):

        Returns:

        """
        if self.is_materialized and tools:
            raise CannotModifyMaterializedAgent(agent_name=self.name)

        for t in tools:
            motley_tool = MotleyTool.from_supported_tool(t)
            if motley_tool.name not in self.tools:
                self.tools[motley_tool.name] = motley_tool

    def as_tool(self, input_schema: Optional[BaseModel] = None) -> MotleyTool:
        """Description

        Args:
            input_schema (:obj:`BaseModel`, optional):

        Returns:
            MotleyTool:
        """

        if not self.description:
            raise ValueError("Agent must have a description to be called as a tool")

        def call_agent(*args, **kwargs):
            # TODO: this thing is hacky, we should have a better way to pass structured input
            if args:
                return self.invoke({"prompt": args[0]})
            if len(kwargs) == 1:
                return self.invoke({"prompt": list(kwargs.values())[0]})
            return self.invoke(kwargs)

        # To be specialized if we expect structured input
        return MotleyTool.from_langchain_tool(
            Tool(
                name=self.name,
                description=self.description,
                func=call_agent,
                args_schema=input_schema,
            )
        )

    # def call_as_tool(self, *args, **kwargs) -> Any:
    #     logger.info("Entering delegation for %s", self.name)
    #     assert self.crew, "can't accept delegated task outside of a crew"
    #
    #     if len(args) > 0:
    #         input_ = args[0]
    #     elif "tool_input" in kwargs:
    #         # Is this a crewai notation?
    #         input_ = kwargs["tool_input"]
    #     else:
    #         input_ = json.dumps(kwargs)
    #
    #     logger.info("Made the args: %s", input_)
    #
    #     # TODO: pass context of parent task to agent nicely?
    #     # TODO: mark the current task as depending on the new task
    #     task = SimpleTaskRecipe(
    #         description=input_,
    #         name=input_,
    #         agent=self,
    #         # TODO inject the new subtask as a dep and reschedule the parent
    #         # TODO probably can't do this from here since we won't know if
    #         # there are other tasks to schedule
    #         crew=self.crew,
    #     )
    #
    #     # TODO: make sure tools return task objects, which are properly used by callers
    #     logger.info("Executing subtask '%s'", task.name)
    #     self.crew.task_graph.set_task_running(task=task)
    #     result = self.crew.execute(task, return_result=True)
    #
    #     logger.info("Finished subtask '%s' - %s", task.name, result)
    #     self.crew.task_graph.set_task_done(task=task)
    #
    #     return result



""" Module description """

from typing import Any, Optional, Sequence

from langchain_core.runnables import RunnableConfig

from motleycrew.agents.parent import MotleyAgentParent
from motleycrew.agents.crewai import CrewAIAgentWithConfig
from motleycrew.common import MotleySupportedTool
from motleycrew.common import MotleyAgentFactory
from motleycrew.tracking import add_default_callbacks_to_langchain_config
from motleycrew.common.utils import ensure_module_is_installed

try:
    from crewai import Task as CrewAI__Task
except ImportError:
    pass


class CrewAIMotleyAgentParent(MotleyAgentParent):
    def __init__(
        self,
        goal: str,
        name: str | None = None,
        agent_factory: MotleyAgentFactory[CrewAIAgentWithConfig] | None = None,
        tools: Sequence[MotleySupportedTool] | None = None,
        verbose: bool = False,
    ):
        """Description

        Args:
            goal (str):
            name (:obj:`str`, optional):
            agent_factory (:obj:`MotleyAgentFactory`, optional):
            tools (:obj:`Sequence[MotleySupportedTool]`, optional:
            verbose (bool):
        """
        ensure_module_is_installed("crewai")
        super().__init__(
            description=goal,
            name=name,
            agent_factory=agent_factory,
            tools=tools,
            verbose=verbose,
        )

    def invoke(
        self,
        task_dict: dict,
        config: Optional[RunnableConfig] = None,
        **kwargs: Any,
    ) -> Any:
        """Description

        Args:
            task_dict (dict):
            config (:obj:`RunnableConfig`, optional):
            **kwargs:

        Returns:
            Any:
        """
        self.materialize()
        prompt = self.compose_prompt(task_dict, task_dict.get("prompt"))

        langchain_tools = [tool.to_langchain_tool() for tool in self.tools.values()]
        config = add_default_callbacks_to_langchain_config(config)

        crewai_task = CrewAI__Task(description=prompt)

        output = self.agent.execute_task(
            task=crewai_task, context=task_dict.get("context"), tools=langchain_tools, config=config
        )
        return output

    # TODO: what do these do?
    def set_cache_handler(self, cache_handler: Any) -> None:
        """Description

        Args:
            cache_handler (Any):

        Returns:
            None:
        """
        return self.agent.set_cache_handler(cache_handler)

    def set_rpm_controller(self, rpm_controller: Any) -> None:
        """Description

        Args:
            rpm_controller (Any):

        Returns:
            None:
        """
        return self.agent.set_rpm_controller(rpm_controller)

    @staticmethod
    def from_agent(
        agent: CrewAIAgentWithConfig,
        tools: Sequence[MotleySupportedTool] | None = None,
        verbose: bool = False,
    ) -> "CrewAIMotleyAgentParent":
        """Description

        Args:
            agent (CrewAIAgentWithConfig):
            tools (:obj:`Sequence[MotleySupportedTool]`, optional):
            verbose (bool):

        Returns:
            CrewAIMotleyAgentParent:
        """
        if tools or agent.tools:
            tools = list(tools or []) + list(agent.tools or [])

        wrapped_agent = CrewAIMotleyAgentParent(
            goal=agent.goal,
            name=agent.role,
            tools=tools,
            verbose=verbose,
        )
        wrapped_agent._agent = agent
        return wrapped_agent



""" Module description

Attributes:
    TaskNodeType (TypeVar):

"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, Sequence, List, Type, TypeVar, Generic, TYPE_CHECKING

from langchain_core.runnables import Runnable
from motleycrew.common.exceptions import TaskDependencyCycleError
from motleycrew.storage import MotleyGraphStore, MotleyGraphNode, MotleyKuzuGraphStore
from motleycrew.tasks import TaskUnitType
from motleycrew.tools import MotleyTool

if TYPE_CHECKING:
    from motleycrew.crew import MotleyCrew


class TaskNode(MotleyGraphNode):
    """Description

    Attributes:
        name (str):
        done (bool):

    """

    __label__ = "TaskNode"
    name: str
    done: bool = False

    def __eq__(self, other):
        return self.id is not None and self.get_label() == other.get_label() and self.id == other.id


TaskNodeType = TypeVar("TaskNodeType", bound=TaskNode)


class Task(ABC, Generic[TaskUnitType]):
    """
    Attributes:
        NODE_CLASS (TaskNodeType):
        TASK_IS_UPSTREAM_LABEL (str):
    """

    NODE_CLASS: Type[TaskNodeType] = TaskNode
    TASK_IS_UPSTREAM_LABEL = "task_is_upstream"

    def __init__(
        self,
        name: str,
        task_unit_class: Type[TaskUnitType],
        crew: Optional[MotleyCrew] = None,
        allow_async_units: bool = False,
    ):
        """Description

        Args:
            name (str):
            task_unit_class (Type[TaskUnitType]):
            crew (:obj:`MotleyCrew`, optional):
            allow_async_units (:obj:'bool', optional)
        """
        self.name = name
        self.done = False
        self.node = self.NODE_CLASS(name=name, done=self.done)
        self.crew = crew
        self.allow_async_units = allow_async_units

        self.task_unit_class = task_unit_class
        self.task_unit_belongs_label = "{}_belongs".format(self.task_unit_class.get_label())

        if crew is not None:
            crew.register_tasks([self])
            self.prepare_graph_store()

    def prepare_graph_store(self):
        """Description

        Returns:

        """
        if isinstance(self.graph_store, MotleyKuzuGraphStore):
            self.graph_store.ensure_node_table(self.NODE_CLASS)
            self.graph_store.ensure_node_table(self.task_unit_class)
            self.graph_store.ensure_relation_table(
                from_class=self.task_unit_class,
                to_class=self.NODE_CLASS,
                label=self.task_unit_belongs_label,
            )

    @property
    def graph_store(self) -> MotleyGraphStore:
        if self.crew is None:
            raise ValueError("Task must be registered with a crew for accessing graph store")
        return self.crew.graph_store

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name}, done={self.done})"

    def __str__(self) -> str:
        return self.__repr__()

    def set_upstream(self, task: Task) -> Task:
        """Description

        Args:
            task (Task):

        Returns:
            Task:
        """
        if self.crew is None or task.crew is None:
            raise ValueError("Both tasks must be registered with a crew")

        if task is self:
            raise TaskDependencyCycleError(f"Task {self.name} can not depend on itself")

        self.crew.add_dependency(upstream=task, downstream=self)

        return self

    def __rshift__(self, other: Task | Sequence[Task]) -> Task:
        if isinstance(other, Task):
            tasks = {other}
        else:
            tasks = other

        for task in tasks:
            task.set_upstream(self)

        return self

    def __rrshift__(self, other: Sequence[Task]) -> Sequence[Task]:
        for task in other:
            self.set_upstream(task)
        return other

    def get_units(self, status: Optional[str] = None) -> List[TaskUnitType]:
        """
        Description

        Args:
            status (str | None): if provided, return only units with this status

        Returns:
            :obj:`list` of :obj:`TaskUnitType`:
        """
        assert self.crew is not None, "Task must be registered with a crew for accessing task units"

        query = (
            "MATCH (unit:{})-[:{}]->(task:{}) WHERE task.id = $self_id"
            + (" AND unit.status = $status" if status is not None else "")
            + " RETURN unit"
        ).format(
            self.task_unit_class.get_label(),
            self.task_unit_belongs_label,
            self.NODE_CLASS.get_label(),
        )

        parameters = {"self_id": self.node.id}
        if status is not None:
            parameters["status"] = status

        task_units = self.graph_store.run_cypher_query(
            query, parameters=parameters, container=self.task_unit_class
        )
        return task_units

    def get_upstream_tasks(self) -> List[Task]:
        """Description

        Returns:
            :obj:`list` of :obj:`Task`
        """
        assert (
            self.crew is not None and self.node.is_inserted
        ), "Task must be registered with a crew for accessing upstream tasks"

        query = (
            "MATCH (upstream)-[:{}]->(downstream:{}) "
            "WHERE downstream.id = $self_id "
            "RETURN upstream"
        ).format(
            self.TASK_IS_UPSTREAM_LABEL,
            self.NODE_CLASS.get_label(),
        )
        upstream_task_nodes = self.graph_store.run_cypher_query(
            query, parameters={"self_id": self.node.id}, container=self.NODE_CLASS
        )
        return [task for task in self.crew.tasks if task.node in upstream_task_nodes]

    def get_downstream_tasks(self) -> List[Task]:
        """Description

        Returns:
            :obj:`list` of :obj:`Task`
        """
        assert (
            self.crew is not None and self.node.is_inserted
        ), "Task must be registered with a crew for accessing downstream tasks"

        query = (
            "MATCH (upstream:{})-[:{}]->(downstream) "
            "WHERE upstream.id = $self_id "
            "RETURN downstream"
        ).format(
            self.NODE_CLASS.get_label(),
            self.TASK_IS_UPSTREAM_LABEL,
        )
        downstream_task_nodes = self.graph_store.run_cypher_query(
            query, parameters={"self_id": self.node.id}, container=self.NODE_CLASS
        )
        return [task for task in self.crew.tasks if task.node in downstream_task_nodes]

    def set_done(self, value: bool = True):
        """Description

        Args:
            value (bool):

        Returns:

        """
        self.done = value
        self.node.done = value

    def register_started_unit(self, unit: TaskUnitType) -> None:
        """Description

        Args:
            unit (TaskUnitType):

        Returns:

        """
        pass

    def register_completed_unit(self, unit: TaskUnitType) -> None:
        """Description

        Args:
            unit (TaskUnitType):

        Returns:

        """
        pass

    @abstractmethod
    def get_next_unit(self) -> TaskUnitType | None:
        """Description

        Returns:
            :obj:`TaskUnitType` | None:
        """
        pass

    @abstractmethod
    def get_worker(self, tools: Optional[List[MotleyTool]]) -> Runnable:
        """Description

        Args:
            tools (:obj:`List[MotleyTool]`, None):

        Returns:
            Runnable:
        """
        pass



""" Module description """
from typing import Any, Optional, List

from langchain_core.runnables import RunnableConfig
from langchain.tools.render import render_text_description

from motleycrew.common.utils import ensure_module_is_installed

try:
    from crewai import Agent
except ImportError:
    Agent = object


class CrewAIAgentWithConfig(Agent):

    def __init__(self, *args, **kwargs):
        """Subclass for CrewAI Agent that overrides the execute_task method to include a config parameter.

        Args:
            *args:
            **kwargs:

        Todo:
            * get rid of this when https://github.com/joaomdmoura/crewAI/pull/483 is merged.
        """
        ensure_module_is_installed("crewai")
        super(CrewAIAgentWithConfig, self).__init__(*args, **kwargs)

    def execute_task(
        self,
        task: Any,
        context: Optional[str] = None,
        tools: Optional[List[Any]] = None,
        config: Optional[RunnableConfig] = None,
    ) -> str:
        """Execute a task with the agent.

        Args:
            task (Any): Task to execute.
            context (:obj:`str`, optional): Context to execute the task in.
            tools (:obj:`List[Any]`, optional): Tools to use for the task.
            config (:obj:`RunnableConfig`, optional): Runnable config.

        Returns:
            Any: Output of the agent
        """
        task_prompt = task.prompt()

        if context:
            task_prompt = self.i18n.slice("task_with_context").format(
                task=task_prompt, context=context
            )

        tools = self._parse_tools(tools or self.tools)
        self.create_agent_executor(tools=tools)
        self.agent_executor.tools = tools
        self.agent_executor.task = task
        self.agent_executor.tools_description = render_text_description(tools)
        self.agent_executor.tools_names = self.__tools_names(tools)

        result = self.agent_executor.invoke(
            {
                "input": task_prompt,
                "tool_names": self.agent_executor.tools_names,
                "tools": self.agent_executor.tools_description,
            },
            config=config,
        )["output"]

        if self.max_rpm:
            self._rpm_controller.stop_rpm_counter()

        return result

    @staticmethod
    def __tools_names(tools) -> str:
        return ", ".join([t.name for t in tools])



""" Project logger configuration module

Attributes:
    logger (logging.Logger): project logger
"""
import logging


# init logger
logger = logging.getLogger("motleycrew")
stream_handler = logging.StreamHandler()
formatter = logging.Formatter(fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)
logger.propagate = False


def configure_logging(verbose: bool = False, debug: bool = False):
    """ Logging configuration

    Args:
        verbose (:obj:`bool`, optional): if true logging level = INFO
        debug (:obj:`bool`, optional): if true logging level = DEBUG else WARNING

    Returns:

    """
    if debug:
        logger.setLevel(logging.DEBUG)
    elif verbose:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)



""" Module description

Attributes:
   PROMPT_TEMPLATE_WITH_DEPS (str):

"""
from __future__ import annotations
from typing import TYPE_CHECKING, Any, Sequence, List, Optional

from motleycrew.tasks.task import Task
from motleycrew.tasks import TaskUnit

from motleycrew.agents.abstract_parent import MotleyAgentAbstractParent
from motleycrew.tools import MotleyTool
from motleycrew.common import logger

if TYPE_CHECKING:
    from motleycrew.crew import MotleyCrew


PROMPT_TEMPLATE_WITH_DEPS = """
{description}

You must use the results of these upstream tasks:

{upstream_results_section}
"""


def compose_simple_task_prompt_with_dependencies(
    description: str, upstream_task_units: List[TaskUnit], default_task_name: str = "Unnamed task"
) -> str:
    """ Description

    Args:
        description (str):
        upstream_task_units (:obj:`list` of :obj:`TaskUnit`):
        default_task_name (:obj:`str`, optional):

    Returns:
        str:
    """
    upstream_results = []
    for unit in upstream_task_units:
        if not unit.output:
            continue

        unit_name = getattr(unit, "name", default_task_name)
        upstream_results.append(f"##{unit_name}\n" + str(unit.output))

    if not upstream_results:
        return description

    upstream_results_section = "\n\n".join(upstream_results)
    return PROMPT_TEMPLATE_WITH_DEPS.format(
        description=description,
        upstream_results_section=upstream_results_section,
    )


class SimpleTaskUnit(TaskUnit):
    """ Description

    Attributes:
        name (str):
        prompt (str):
        message_history (:obj:`list` of :obj:`str`):

    """
    name: str
    prompt: str
    message_history: List[str] = []


class SimpleTask(Task):
    def __init__(
        self,
        crew: MotleyCrew,
        description: str,
        name: str | None = None,
        agent: MotleyAgentAbstractParent | None = None,
        tools: Sequence[MotleyTool] | None = None,
        documents: Sequence[Any] | None = None,
        creator_name: str | None = None,
        return_to_creator: bool = False,
    ):
        """ Description

        Args:
            crew (MotleyCrew):
            description (str):
            name (:obj:`str`, optional):
            agent (:obj:`MotleyAgentAbstractParent`, optional):
            tools (:obj:`Sequence[MotleyTool]`, optional):
            documents (:obj:`Sequence[Any]`, optional):
            creator_name (:obj:`str`, optional):
            return_to_creator (:obj:`bool`, optional):
        """
        super().__init__(name=name or description, task_unit_class=SimpleTaskUnit, crew=crew)
        self.description = description
        self.agent = agent  # to be auto-assigned at crew creation if missing?
        self.tools = tools or []
        # should tasks own agents or should agents own tasks?
        self.documents = documents  # to be passed to an auto-init'd retrieval, later on
        self.creator_name = creator_name or "Human"
        self.return_to_creator = (
            return_to_creator  # for orchestrator to know to send back to creator
        )
        self.output = None  # to be filled in by the agent(s) once the task is complete

    def register_completed_unit(self, unit: SimpleTaskUnit) -> None:
        """ Description

        Args:
            unit (SimpleTaskUnit):

        Returns:

        """
        assert isinstance(unit, SimpleTaskUnit)
        assert unit.done

        self.output = unit.output
        self.set_done()

    def get_next_unit(self) -> SimpleTaskUnit | None:
        """ Description

        Returns:
            :obj:`SimpleTaskUnit`, None:
        """
        if self.done:
            logger.info("Task %s is already done", self)
            return None

        upstream_tasks = self.get_upstream_tasks()
        if not all(task.done for task in upstream_tasks):
            return None

        upstream_task_units = [unit for task in upstream_tasks for unit in task.get_units()]
        prompt = compose_simple_task_prompt_with_dependencies(self.description, upstream_task_units)
        return SimpleTaskUnit(
            name=self.name,
            prompt=prompt,
        )

    def get_worker(self, tools: Optional[List[MotleyTool]]) -> MotleyAgentAbstractParent:
        """ Description

        Args:
            tools (:obj:`List[MotleyTool]`, :obj:`None`):

        Returns:
            MotleyAgentAbstractParent
        """
        if self.crew is None:
            raise ValueError("Task is not associated with a crew")
        if self.agent is None:
            raise ValueError("Task is not associated with an agent")

        if hasattr(self.agent, "is_materialized") and self.agent.is_materialized and tools:
            logger.warning(
                "Agent %s is already materialized, can't add extra tools %s", self.agent, tools
            )

        if hasattr(self.agent, "add_tools") and tools:
            logger.info("Adding tools %s to agent %s", tools, self.agent)
            self.agent.add_tools(tools)

        # TODO: that's a pretty big assumption on agent structure. Necessary?
        self.agent.crew = self.crew

        return self.agent



""" Module description """
from typing import Optional, Any, Sequence

from motleycrew.tools import MotleyTool
from motleycrew.common import MotleySupportedTool
from motleycrew.common import LLMFramework
from motleycrew.common.llms import init_llm
from motleycrew.agents.abstract_parent import MotleyAgentAbstractParent
from motleycrew.agents.crewai import CrewAIMotleyAgentParent
from motleycrew.agents.crewai import CrewAIAgentWithConfig


class CrewAIMotleyAgent(CrewAIMotleyAgentParent):
    def __init__(
        self,
        role: str,
        goal: str,
        backstory: str,
        delegation: bool = False,
        tools: Sequence[MotleySupportedTool] | None = None,
        llm: Optional[Any] = None,
        verbose: bool = False,
    ):
        """ Description

        Args:
            role (str):
            goal (str):
            backstory (str):
            delegation (bool):
            tools (:obj:`Sequence[MotleySupportedTool]`, optional):
            llm (:obj:'Any', optional):
            verbose (bool):
        """
        if tools is None:
            tools = []

        if llm is None:
            # CrewAI uses Langchain LLMs by default
            llm = init_llm(llm_framework=LLMFramework.LANGCHAIN)

        if delegation:
            raise ValueError(
                "'delegation' is not supported, pass the agents you want to delegate to as tools instead."
            )

        def agent_factory(tools: dict[str, MotleyTool]):
            langchain_tools = [t.to_langchain_tool() for t in tools.values()]
            agent = CrewAIAgentWithConfig(
                role=role,
                goal=goal,
                backstory=backstory,
                verbose=verbose,
                allow_delegation=False,
                tools=langchain_tools,
                llm=llm,
            )
            return agent

        super().__init__(
            goal=goal,
            name=role,
            agent_factory=agent_factory,
            tools=tools,
            verbose=verbose,
        )



""" Module description """

from typing import Any, Optional, Sequence, Callable, Union

from langchain.agents import AgentExecutor
from langchain.agents.agent import RunnableAgent
from langchain_core.agents import AgentFinish, AgentAction
from langchain_core.runnables import Runnable, RunnableConfig, RunnableLambda
from langchain_core.runnables.history import RunnableWithMessageHistory, GetSessionHistoryCallable
from langchain_core.chat_history import InMemoryChatMessageHistory
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import BasePromptTemplate
from langchain_core.tools import BaseTool
from langchain_core.tools import Tool


from motleycrew.agents.parent import MotleyAgentParent

from motleycrew.tools import MotleyTool
from motleycrew.tracking import add_default_callbacks_to_langchain_config
from motleycrew.common import MotleySupportedTool, logger
from motleycrew.common import MotleyAgentFactory
from motleycrew.common import LLMFramework
from motleycrew.common.llms import init_llm


class LangchainMotleyAgent(MotleyAgentParent):
    def __init__(
        self,
        description: str | None = None,
        name: str | None = None,
        agent_factory: MotleyAgentFactory[AgentExecutor] | None = None,
        tools: Sequence[MotleySupportedTool] | None = None,
        output_handler: MotleySupportedTool | None = None,
        verbose: bool = False,
        chat_history: bool | GetSessionHistoryCallable = True,
    ):
        """Description

        Args:
            description (:obj:`str`, optional):
            name (:obj:`str`, optional):
            agent_factory (:obj:`MotleyAgentFactory`, optional):
            tools (:obj:`Sequence[MotleySupportedTool]`, optional):
            output_handler (:obj:`MotleySupportedTool`, optional):
            verbose (bool):
            chat_history (:obj:`bool`, :obj:`GetSessionHistoryCallable`):
            Whether to use chat history or not. If `True`, uses `InMemoryChatMessageHistory`.
            If a callable is passed, it is used to get the chat history by session_id.
            See Langchain `RunnableWithMessageHistory` get_session_history param for more details.
        """
        super().__init__(
            description=description,
            name=name,
            agent_factory=agent_factory,
            tools=tools,
            output_handler=output_handler,
            verbose=verbose,
        )

        if chat_history is True:
            chat_history = InMemoryChatMessageHistory()
            self.get_session_history_callable = lambda _: chat_history
        else:
            self.get_session_history_callable = chat_history

        self._agent_finish_blocker_tool = self._create_agent_finish_blocker_tool()

    def _create_agent_finish_blocker_tool(self) -> BaseTool:
        """Create a tool that will force the agent to retry if it attempts to return the output
        bypassing the output handler.
        """

        def create_agent_finish_blocking_message(input: Any) -> str:
            return f"{input}\n\nYou must use {self.output_handler.name} to return the final output."

        return Tool.from_function(
            name="agent_finish_blocker",
            description="",
            func=create_agent_finish_blocking_message,
        )

    def _block_agent_finish(self, input: Any):
        """Intercept AgentFinish for forcing output via output handler.
        If the agent attempts to return the output bypassing the output handler,
        a tool call to the agent_finish_blocker_tool will be made
        so that one more AgentExecutor iteration is forced.
        """
        if isinstance(input, AgentFinish) and self.output_handler:
            return [
                AgentAction(
                    tool=self._agent_finish_blocker_tool.name,
                    tool_input={"input": input.return_values},
                    log="\nDetected AgentFinish, blocking it to force output via output handler.\n",
                )
            ]
        return input

    def materialize(self):
        """Materialize the agent and wrap it in RunnableWithMessageHistory if needed."""
        if self.is_materialized:
            return

        super().materialize()
        assert isinstance(self._agent, AgentExecutor)

        if self.output_handler:
            self._agent.tools += [self._agent_finish_blocker_tool]

            if isinstance(self._agent.agent, Runnable):
                logger.info("Patching agent to force output via output handler")
                agent_with_forced_output_via_handler = self._agent.agent | RunnableLambda(
                    self._block_agent_finish
                )
                self._agent.agent = agent_with_forced_output_via_handler

            elif isinstance(self._agent.agent, RunnableAgent):
                logger.info("Patching agent to force output via output handler")
                agent_with_forced_output_via_handler = self._agent.agent.runnable | RunnableLambda(
                    self._block_agent_finish
                )
                self._agent.agent.runnable = agent_with_forced_output_via_handler

            else:
                logger.warning(
                    "Agent is not a Runnable, cannot patch it to force output via output handler"
                )

        if self.get_session_history_callable:
            logger.info("Wrapping agent in RunnableWithMessageHistory")

            if isinstance(self._agent, RunnableWithMessageHistory):
                return
            self._agent = RunnableWithMessageHistory(
                runnable=self._agent,
                get_session_history=self.get_session_history_callable,
                input_messages_key="input",
                history_messages_key="chat_history",
            )

    def invoke(
        self,
        task_dict: dict,
        config: Optional[RunnableConfig] = None,
        **kwargs: Any,
    ) -> Any:
        """Description

        Args:
            task_dict (dict):
            config (:obj:`RunnableConfig`, optional):
            **kwargs:

        Returns:

        """
        self.materialize()
        prompt = self.compose_prompt(task_dict, task_dict.get("prompt"))

        config = add_default_callbacks_to_langchain_config(config)
        if self.get_session_history_callable:
            config["configurable"] = config.get("configurable") or {}
            config["configurable"]["session_id"] = (
                config["configurable"].get("session_id") or "default"
            )

        caught_direct_output, output = self._run_and_catch_output(
            action=self.agent.invoke,
            action_args=(
                {"input": prompt},
                config,
            ),
            action_kwargs=kwargs,
        )

        if not caught_direct_output:
            output = output.get("output")  # unpack native Langchain output

        return output

    @staticmethod
    def from_agent(
        agent: AgentExecutor,
        goal: str,
        tools: Sequence[MotleySupportedTool] | None = None,
        verbose: bool = False,
    ) -> "LangchainMotleyAgent":
        """Description

        Args:
            agent (AgentExecutor):
            goal (str):
            tools(:obj:`Sequence[MotleySupportedTool]`, optional):
            verbose (bool):

        Returns:
            LangchainMotleyAgent
        """
        # TODO: do we really need to unite the tools implicitly like this?
        # TODO: confused users might pass tools both ways at the same time
        # TODO: and we will silently unite them, which can have side effects (e.g. doubled tools)
        # TODO: besides, getting tools can be tricky for other frameworks (e.g. LlamaIndex)
        if tools or agent.tools:
            tools = list(tools or []) + list(agent.tools or [])

        wrapped_agent = LangchainMotleyAgent(description=goal, tools=tools, verbose=verbose)
        wrapped_agent._agent = agent
        return wrapped_agent



""" Module description """
# https://nodejs.org/en/download
# npm install -g @mermaid-js/mermaid-cli
import os.path
import subprocess
import io
import tempfile
from typing import Optional

from langchain_core.pydantic_v1 import create_model, Field
from langchain_core.tools import Tool
from motleycrew.tools import MotleyTool


class MermaidEvaluatorTool(MotleyTool):
    def __init__(self, format: Optional[str] = "svg"):
        """ Description

        Args:
            format (:obj:`str`, None):
        """
        def eval_mermaid_partial(mermaid_code: str):
            return eval_mermaid(mermaid_code, format)

        langchain_tool = Tool.from_function(
            func=eval_mermaid_partial,
            name="mermaid_evaluator_tool",
            description="Evaluates Mermaid code and returns the output as a BytesIO object.",
            args_schema=create_model(
                "MermaidEvaluatorToolInput",
                mermaid_code=(str, Field(description="The Mermaid code to evaluate.")),
            ),
        )
        super().__init__(langchain_tool)


def eval_mermaid(mermaid_code: str, format: Optional[str] = "svg") -> io.BytesIO:
    """ Description

    Args:
        mermaid_code (str):
        format (:obj:`str`, optional):

    Returns:
        io.BytesIO:
    """
    with tempfile.NamedTemporaryFile(delete=True, mode="w+", suffix=".mmd") as temp_in:
        temp_in.write(mermaid_code)
        temp_in.flush()  # Ensure all data is written to disk

        if format in ["md", "markdown"]:
            raise NotImplementedError("Markdown format is not yet supported in this wrapper.")
        assert format in [
            "svg",
            "png",
            "pdf",
        ], "Invalid format specified, must be svg, png, or pdf."
        out_file = f"output.{format}"

        # Prepare the command to call the mermaid CLI
        full_code = f"mmdc -i {temp_in.name} -o {out_file} -b transparent"

        try:
            # Execute the command
            subprocess.run(
                full_code,
                shell=True,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            # If the process succeeds, read the output file into BytesIO
            with open(out_file, "rb") as f:
                output_bytes = io.BytesIO(f.read())
            return output_bytes
        except subprocess.CalledProcessError as e:
            # If the process fails, print the error message
            return e.stderr.decode()
        finally:
            # Clean up the output file if it exists
            try:
                os.remove(out_file)
            except FileNotFoundError:
                pass


# Define the Mermaid code for the flowchart

if __name__ == "__main__":
    mermaid_code = """
    graph TD;
        A[Start] --> B[Decision]
        B -- Yes --> C[Option 1]
        B -- No --> D[Option 2]
        C --> E[End]
        D --> E
        E[End] --> F[End]
        
        [[
    """

    out1 = eval_mermaid(mermaid_code)
    output_file_path = "output_file.bin"
    if isinstance(out1, str):
        print(out1)
        exit(1)
    # Ensure the pointer is at the beginning of the BytesIO object
    out1.seek(0)

    # Open the output file in binary write mode and write the contents of the BytesIO object
    with open(output_file_path, "wb") as file_output:
        file_output.write(out1.read())
    tool = MermaidEvaluatorTool()
    out2 = tool.invoke({"mermaid_code": mermaid_code})
    print(out2)



"""
Code derived from: https://github.com/run-llama/llama_index/blob/802064aee72b03ab38ead0cda780cfa3e37ce728/llama-index-integrations/graph_stores/llama-index-graph-stores-kuzu/llama_index/graph_stores/kuzu/base.py
Kùzu graph store index.
"""

from typing import Any, Dict, List, Optional, Type, TypeVar

from kuzu import Connection, PreparedStatement, QueryResult
import json
import os

from motleycrew.storage import MotleyGraphStore
from motleycrew.storage import MotleyGraphNode
from motleycrew.storage import MotleyGraphNodeType
from motleycrew.common import logger


class MotleyKuzuGraphStore(MotleyGraphStore):
    """
    Attributes:
        ID_ATTR (str): _id
        JSON_CONTENT_PREFIX (str): "JSON__"
        PYTHON_TO_CYPHER_TYPES_MAPPING (dict)

    """
    ID_ATTR = "_id"

    JSON_CONTENT_PREFIX = "JSON__"

    PYTHON_TO_CYPHER_TYPES_MAPPING = {
        int: "INT64",  # TODO: enforce size when creating and updating nodes and relations
        Optional[int]: "INT64",
        str: "STRING",
        Optional[str]: "STRING",
        float: "DOUBLE",
        Optional[float]: "DOUBLE",
        bool: "BOOLEAN",
        Optional[bool]: "BOOLEAN",
    }

    def __init__(self, database: Any) -> None:
        """ Description

        Args:
            database (Any):
        """
        self.database = database
        self.connection = Connection(database)
        # Workaround for Kuzu requiring at least one relation table
        # TODO: fix, https://github.com/kuzudb/kuzu/issues/3488
        self.ensure_node_table(MotleyGraphNode)
        self.ensure_relation_table(MotleyGraphNode, MotleyGraphNode, "dummy")

    def __repr__(self):
        return f"{self.__class__.__name__}(path={self.database_path})"

    def __str__(self):
        return self.__repr__()

    @property
    def database_path(self) -> str:
        return os.path.abspath(self.database.database_path)

    def _execute_query(
        self, query: str | PreparedStatement, parameters: Optional[dict[str, Any]] = None
    ) -> QueryResult:
        """ Execute a query, logging it for debugging purposes

        Args:
            query (:obj:`str`, :obj:`PreparedStatement`):
            parameters (:obj:`dict` of :obj:`str` , :obj:`Any`):

        Returns:
            QueryResult:
        """
        logger.debug("Executing query: %s", query)
        if parameters:
            logger.debug("with parameters: %s", parameters)

        # TODO: retries?
        return self.connection.execute(query=query, parameters=parameters)

    def _check_node_table_exists(self, label: str):
        """ Description

        Args:
            label (str):

        Returns:

        """
        return label in self.connection._get_node_table_names()

    def _check_rel_table_exists(
        self,
        from_label: Optional[str] = None,
        to_label: Optional[str] = None,
        rel_label: Optional[str] = None,
    ):
        """ Description

        Args:
            from_label (:obj:`str`, optional):
            to_label (:obj:`str`, optional):
            rel_label (:obj:`str`, optional):

        Returns:

        """
        for row in self.connection._get_rel_table_names():
            if (
                (rel_label is None or row["name"] == rel_label)
                and (from_label is None or row["src"] == from_label)
                and (to_label is None or row["dst"] == to_label)
            ):
                return True
        return False

    def _get_node_property_names(self, label: str):
        """ Description

        Args:
            label (str):

        Returns:

        """
        return self.connection._get_node_property_names(table_name=label)

    def ensure_node_table(self, node_class: Type[MotleyGraphNode]) -> str:
        """ Create a table for storing nodes of that class if such does not already exist.
        If it does exist, create all missing columns.

         Args:
            node_class (Type[MotleyGraphNode]):
        Returns:
            str: Table name

        """
        table_name = node_class.get_label()
        if not self._check_node_table_exists(table_name):
            logger.info("Node table %s does not exist in the database, creating", table_name)
            self._execute_query(
                "CREATE NODE TABLE {} (id SERIAL, PRIMARY KEY(id))".format(table_name)
            )

        # Create missing property columns
        existing_property_names = self._get_node_property_names(node_class.get_label())
        for field_name, field in node_class.model_fields.items():
            if field_name not in existing_property_names:
                logger.info(
                    "Property %s not present in table for label %s, creating",
                    field_name,
                    node_class.get_label(),
                )
                cypher_type, is_json = (
                    MotleyKuzuGraphStore._get_cypher_type_and_is_json_by_python_type_annotation(
                        field.annotation
                    )
                )

                self._execute_query(
                    "ALTER TABLE {} ADD {} {}".format(table_name, field_name, cypher_type)
                )
        return table_name

    def ensure_relation_table(
        self, from_class: Type[MotleyGraphNode], to_class: Type[MotleyGraphNode], label: str
    ):
        """ Create a table for storing relations from from_node-like nodes to to_node-like nodes,
        if such does not already exist.

        Args:
            from_class (Type[MotleyGraphNode]):
            to_class (Type[MotleyGraphNode]):
            label (str):

        Returns:

        """
        if not self._check_rel_table_exists(
            from_label=from_class.get_label(), to_label=to_class.get_label(), rel_label=label
        ):
            logger.info(
                "Relation table %s from %s to %s does not exist in the database, creating",
                label,
                from_class.get_label(),
                to_class.get_label(),
            )

            self._execute_query(
                "CREATE REL TABLE {} (FROM {} TO {})".format(
                    label, from_class.get_label(), to_class.get_label()
                )
            )

    def check_node_exists_by_class_and_id(
        self, node_class: Type[MotleyGraphNode], node_id: int
    ) -> bool:
        """ Check if a node of given class with given id is present in the database.

        Args:
            node_class (Type[MotleyGraphNode]):
            node_id (int):

        Returns:
            bool:
        """
        if not self._check_node_table_exists(node_class.get_label()):
            return False

        is_exists_result = self._execute_query(
            "MATCH (n:{}) WHERE n.id = $node_id RETURN n.id".format(node_class.get_label()),
            {"node_id": node_id},
        )
        return is_exists_result.has_next()

    def check_node_exists(self, node: MotleyGraphNode) -> bool:
        """ Check if the given node is present in the database.

        Args:
            node (MotleyGraphNode):

        Returns:
            bool:
        """
        if node.id is None:
            return False  # for cases when id attribute is not set => node does not exist

        return self.check_node_exists_by_class_and_id(node_class=node.__class__, node_id=node.id)

    def check_relation_exists(
        self, from_node: MotleyGraphNode, to_node: MotleyGraphNode, label: Optional[str] = None
    ) -> bool:
        """ Check if a relation exists between two nodes with given label.

        Args:
            from_node (MotleyGraphNode):
            to_node (MotleyGraphNode):
            label (:obj:`str`, None):

        Returns:
            bool:
        """
        if from_node.id is None or to_node.id is None:
            return False

        if (
            not self._check_node_table_exists(from_node.get_label())
            or not self._check_node_table_exists(to_node.get_label())
            or not self._check_rel_table_exists(
                from_label=from_node.get_label(), to_label=to_node.get_label(), rel_label=label
            )
        ):
            return False

        query = (
            "MATCH (n1:{})-[r{}]->(n2:{}) "
            "WHERE n1.id = $from_node_id AND n2.id = $to_node_id "
            "RETURN r".format(
                from_node.get_label(),
                (":" + label) if label else "",
                to_node.get_label(),
            )
        )
        parameters = {
            "from_node_id": from_node.id,
            "to_node_id": to_node.id,
        }

        is_exists_result = self._execute_query(query=query, parameters=parameters)
        return is_exists_result.has_next()

    def get_node_by_class_and_id(
        self, node_class: Type[MotleyGraphNodeType], node_id: int
    ) -> Optional[MotleyGraphNodeType]:
        """ Retrieve the node of given class with given id if it is present in the database.
        Otherwise, return None.

        Args:
            node_class (Type[MotleyGraphNodeType]):
            node_id (int):

        Returns:
            :obj:`MotleyGraphNodeType`, None:
        """
        if not self._check_node_table_exists(node_class.get_label()):
            return None

        query = """
                    MATCH (n:{})
                    WHERE n.id = $node_id
                    RETURN n;
                """.format(
            node_class.get_label()
        )
        query_result = self._execute_query(query, {"node_id": node_id})

        if query_result.has_next():
            row = query_result.get_next()
            return self._deserialize_node(node_dict=row[0], node_class=node_class)

    def insert_node(self, node: MotleyGraphNodeType) -> MotleyGraphNodeType:
        """ Insert a new node and populate its id.
        If node table or some columns do not exist, this method also creates them.

        Args:
            node (MotleyGraphNodeType):

        Returns:
            MotleyGraphNodeType
        """
        assert node.id is None, "Entity has its id set, looks like it is already in the DB"

        self.ensure_node_table(type(node))
        logger.info("Inserting new node with label %s: %s", node.get_label(), node)

        cypher_mapping, parameters = MotleyKuzuGraphStore._node_to_cypher_mapping_with_parameters(
            node
        )
        create_result = self._execute_query(
            "CREATE (n:{} {}) RETURN n".format(node.get_label(), cypher_mapping),
            parameters=parameters,
        )
        assert create_result.has_next()
        logger.info("Node created OK")

        created_object = create_result.get_next()[0]
        created_object_id = created_object.get("id")
        assert created_object_id is not None, "BUG: created object ID was not returned: {}".format(
            created_object
        )

        MotleyKuzuGraphStore._set_node_id(node=node, node_id=created_object_id)
        node.__graph_store__ = self
        return node

    def create_relation(
        self, from_node: MotleyGraphNode, to_node: MotleyGraphNode, label: str
    ) -> None:
        """ Create a relation between existing nodes.
        If relation table does not exist, this method also creates them

        Args:
            from_node (MotleyGraphNode):
            to_node (MotleyGraphNode):
            label (str):

        Returns:

        """
        assert self.check_node_exists(from_node), (
            "From-node is not present in the database, "
            "consider using upsert_triplet() for such cases"
        )
        assert self.check_node_exists(to_node), (
            "To-node is not present in the database, "
            "consider using upsert_triplet() for such cases"
        )

        self.ensure_relation_table(from_class=type(from_node), to_class=type(to_node), label=label)

        logger.info(
            "Creating relation %s from %s:%s to %s:%s",
            label,
            from_node.get_label(),
            from_node.id,
            to_node.get_label(),
            to_node.id,
        )

        create_result = self._execute_query(
            (
                "MATCH (n1:{}), (n2:{}) WHERE n1.id = $from_id AND n2.id = $to_id "
                "CREATE (n1)-[r:{}]->(n2) "
                "RETURN r"
            ).format(from_node.get_label(), to_node.get_label(), label),
            {
                "from_id": from_node.id,
                "to_id": to_node.id,
            },
        )
        assert create_result.has_next()
        logger.info("Relation created OK")

    def upsert_triplet(self, from_node: MotleyGraphNode, to_node: MotleyGraphNode, label: str):
        """ Create a relation with a given label between nodes, if such does not already exist.
        If the nodes do not already exist, create them too.
        This method also creates and/or updates all necessary tables

        Args:
            from_node (MotleyGraphNode):
            to_node (MotleyGraphNode):
            label (str):

        Returns:

        """
        if not self.check_node_exists(from_node):
            logger.info("Node %s does not exist, creating", from_node)
            self.insert_node(from_node)

        if not self.check_node_exists(to_node):
            logger.info("Node %s does not exist, creating", to_node)
            self.insert_node(to_node)

        if not self.check_relation_exists(from_node=from_node, to_node=to_node, label=label):
            logger.info("Relation from %s to %s does not exist, creating", from_node, to_node)
            self.create_relation(from_node=from_node, to_node=to_node, label=label)

    def delete_node(self, node: MotleyGraphNode) -> None:
        """ Delete a given node and its relations.

        Args:
            node (MotleyGraphNode):

        Returns:

        """

        def inner_delete_relations(node_label: str, node_id: int) -> None:
            if not self.connection._get_rel_table_names():
                # Avoid Kuzu error when no relation tables exist in the database
                return

            # Undirected relation removal is not supported for some reason
            if self._check_rel_table_exists(from_label=node_label):
                self._execute_query(
                    "MATCH (n:{})-[r]->() WHERE n.id = $node_id DELETE r".format(node_label),
                    {"node_id": node_id},
                )
            if self._check_rel_table_exists(to_label=node_label):
                self._execute_query(
                    "MATCH (n:{})<-[r]-() WHERE n.id = $node_id DELETE r".format(node_label),
                    {"node_id": node_id},
                )

        def inner_delete_node(node_label: str, node_id: int) -> None:
            self._execute_query(
                "MATCH (n:{}) WHERE n.id = $node_id DELETE n".format(node_label),
                {"node_id": node_id},
            )

        assert self.check_node_exists(node), "Cannot delete nonexistent node: {}".format(node)

        inner_delete_relations(node_label=node.get_label(), node_id=node.id)
        inner_delete_node(node_label=node.get_label(), node_id=node.id)

        MotleyKuzuGraphStore._set_node_id(node, None)

    def update_property(self, node: MotleyGraphNode, property_name: str) -> MotleyGraphNode:
        """ Update a graph node's property with the corresponding value from the node object.

        Args:
            node (MotleyGraphNode):
            property_name (str):

        Returns:

        """
        property_value = getattr(node, property_name)
        if property_value is None:
            # TODO: remove after updating Kuzu to v0.3.3 (https://github.com/kuzudb/kuzu/pull/3098)
            raise Exception("Kuzu does not support NoneType parameters for properties for now")

        existing_property_names = self._get_node_property_names(node.get_label())

        assert property_name in node.model_fields, "No such field in node model {}: {}".format(
            node.__class__.__name__, property_name
        )

        assert self.check_node_exists(node)
        assert property_name in existing_property_names, "No such field in DB table {}: {}".format(
            node.get_label(), property_name
        )

        _, is_json = MotleyKuzuGraphStore._get_cypher_type_and_is_json_by_python_type_annotation(
            node.model_fields[property_name].annotation
        )

        db_property_name = property_name
        if is_json:
            db_property_value = MotleyKuzuGraphStore.JSON_CONTENT_PREFIX + json.dumps(
                property_value
            )
        else:
            db_property_value = property_value

        query = """
                    MATCH (n:{})
                    WHERE n.id = $node_id
                    SET n.{} = $property_value RETURN n;
                """.format(
            node.get_label(), db_property_name
        )

        query_result = self._execute_query(
            query,
            {"node_id": node.id, "property_value": db_property_value},
        )
        assert query_result.has_next()
        row = query_result.get_next()
        node_dict = row[0]
        assert node_dict[db_property_name] == db_property_value

        return node

    def run_cypher_query(
        self,
        query: str,
        parameters: Optional[dict] = None,
        container: Optional[Type[MotleyGraphNodeType]] = None,
    ) -> list[list | MotleyGraphNodeType]:
        """ Run a Cypher query and return the results.
        If container class is provided, deserialize the results into objects of that class.

        Args:
            query (:obj:`dict`, None):
            parameters (:obj:`dict`, optional):
            container (:obj:`Type[MotleyGraphNodeType]`, optional):

        Returns:

        """
        query_result = self._execute_query(query=query, parameters=parameters)
        retval = []
        while query_result.has_next():
            row = query_result.get_next()
            if container is not None:
                assert len(row) == 1, "Expected single column result for deserialization"
                retval.append(self._deserialize_node(node_dict=row[0], node_class=container))
            else:
                retval.append(row)
        return retval

    def _deserialize_node(
        self, node_dict: dict, node_class: Type[MotleyGraphNode]
    ) -> MotleyGraphNode:
        """ Description

        Args:
            node_dict (dict):
            node_class (Type[MotleyGraphNode]):

        Returns:
            MotleyGraphNode
        """
        for field_name, value in node_dict.copy().items():
            if isinstance(value, str) and value.startswith(
                MotleyKuzuGraphStore.JSON_CONTENT_PREFIX
            ):
                logger.debug(
                    "Value for field %s is marked as JSON, attempting to deserialize: %s",
                    field_name,
                    value,
                )
                node_dict[field_name] = json.loads(
                    value[len(MotleyKuzuGraphStore.JSON_CONTENT_PREFIX) :]
                )

        node = node_class.model_validate(node_dict)
        node._id = node_dict["id"]
        node.__graph_store__ = self
        if "id" in node_dict:
            MotleyKuzuGraphStore._set_node_id(node, node_dict["id"])
        return node

    @staticmethod
    def _set_node_id(node: MotleyGraphNode, node_id: Optional[int]) -> None:
        """ Description

        Args:
            node (MotleyGraphNode):
            node_id (:obj:`int`, optional):

        Returns:

        """
        setattr(node, MotleyKuzuGraphStore.ID_ATTR, node_id)

    @staticmethod
    def _node_to_cypher_mapping_with_parameters(node: MotleyGraphNode) -> tuple[str, dict]:
        """ Description

        Args:
            node (MotleyGraphNode):

        Returns:
            :obj:`tuple` of :obj:`str`, :obj:`dict`:
        """
        node_dict = node.model_dump()

        parameters = {}

        cypher_mapping = "{"
        for field_name, value in node_dict.items():
            assert field_name != "id", "id field is reserved for node id"
            if value is None:
                # TODO: remove after updating Kuzu to v0.3.3
                # (https://github.com/kuzudb/kuzu/pull/3098)
                continue

            _, is_json = (
                MotleyKuzuGraphStore._get_cypher_type_and_is_json_by_python_type_annotation(
                    node.model_fields[field_name].annotation
                )
            )
            if is_json and value is not None:
                value = json.dumps(value)
                value = MotleyKuzuGraphStore.JSON_CONTENT_PREFIX + value

            cypher_mapping += f"{field_name}: ${field_name}, "
            parameters[field_name] = value

        cypher_mapping = cypher_mapping.rstrip(", ") + "}"
        return cypher_mapping, parameters

    @staticmethod
    def _get_cypher_type_and_is_json_by_python_type_annotation(
        annotation: Type,
    ) -> tuple[str, bool]:
        """ Determine suitable Cypher data type by Python/Pydantic type annotation,
        and whether the data should be stored in JSON-serialized strings.

        Args:
            annotation (Type):

        Returns:
            :obj:`tuple` of :obj:`str`, :obj:`bool`:
        """
        cypher_type = MotleyKuzuGraphStore.PYTHON_TO_CYPHER_TYPES_MAPPING.get(annotation)
        if not cypher_type:
            logger.warning(
                "No known Cypher type matching annotation %s, will use JSON string",
                annotation,
            )
            return MotleyKuzuGraphStore.PYTHON_TO_CYPHER_TYPES_MAPPING[str], True
        return cypher_type, False

    @classmethod
    def from_persist_dir(
        cls,
        persist_dir: str,
    ) -> "MotleyKuzuGraphStore":
        """ Load from persist dir.

        Args:
            persist_dir (str):

        Returns:
            MotleyKuzuGraphStore:
        """
        try:
            import kuzu
        except ImportError:
            raise ImportError("Please install kuzu: pip install kuzu")
        database = kuzu.Database(persist_dir)
        return cls(database)

    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any]) -> "MotleyKuzuGraphStore":
        """Initialize graph store from configuration dictionary.

        Args:
            config_dict (dict): Configuration dictionary.

        Returns:
            MotleyKuzuGraphStore:Graph store.
        """
        return cls(**config_dict)


if __name__ == "__main__":
    from pathlib import Path
    import shutil

    here = Path(__file__).parent
    db_path = here / "test1"
    shutil.rmtree(db_path, ignore_errors=True)
    db = kuzu.Database(str(db_path))
    graph_store = MotleyKuzuGraphStore(db)

    class Question(MotleyGraphNode):
        question: str
        answer: Optional[str] = None
        context: Optional[List[str]] = None

    q1 = Question(question="q1")
    graph_store.insert_node(q1)
    assert getattr(q1, "_id", None) is not None
    q1_id = q1._id

    assert graph_store.check_node_exists(q1)
    assert graph_store.check_node_exists_by_class_and_id(node_class=Question, node_id=q1_id)

    q2 = Question(question="q2", answer="a2")
    graph_store.upsert_triplet(from_node=q1, to_node=q2, label="p")
    assert getattr(q2, "_id", None) is not None
    q2_id = q2._id

    assert graph_store.check_relation_exists(from_node=q1, to_node=q2, label="p")
    assert not graph_store.check_relation_exists(from_node=q2, to_node=q1)

    graph_store.delete_node(q1)
    assert not graph_store.check_node_exists(q1)
    assert graph_store.get_node_by_class_and_id(node_class=Question, node_id=q1_id) is None

    q2.context = ["abc", "def"]
    assert graph_store.get_node_by_class_and_id(node_class=Question, node_id=q2_id).context == [
        "abc",
        "def",
    ]

    print(f"docker run -p 8000:8000  -v {db_path}:/database --rm kuzudb/explorer: latest")
    print("MATCH (A)-[r]->(B) RETURN *;")



""" Module description"""

from typing import Sequence, Optional

from langchain import hub
from langchain_core.language_models import BaseLanguageModel
from langchain_core.runnables.history import GetSessionHistoryCallable
from langchain.agents import AgentExecutor
from langchain.agents import create_react_agent

from motleycrew.agents.langchain import LangchainMotleyAgent
from motleycrew.tools import MotleyTool
from motleycrew.common import MotleySupportedTool
from motleycrew.common import LLMFramework
from motleycrew.common.llms import init_llm


OUTPUT_HANDLER_WITH_DEFAULT_PROMPT_MESSAGE = (
    "Langchain's default ReAct prompt tells the agent to include a final answer keyword, "
    "which later confuses the parser when an output handler is used. "
    "Please provide a custom prompt if using an output handler."
)


class ReActMotleyAgent(LangchainMotleyAgent):
    def __init__(
        self,
        tools: Sequence[MotleySupportedTool],
        description: str | None = None,
        name: str | None = None,
        prompt: str | None = None,
        output_handler: MotleySupportedTool | None = None,
        chat_history: bool | GetSessionHistoryCallable = True,
        handle_parsing_errors: bool = True,
        handle_tool_errors: bool = True,
        llm: BaseLanguageModel | None = None,
        verbose: bool = False,
    ):
        """Basic ReAct agent compatible with older models without dedicated tool calling support.
        It's probably better to use the more advanced `ReActToolCallingAgent` with newer models.

        Args:
            tools (Sequence[MotleySupportedTool]):
            description (:obj:`str`, optional):
            name (:obj:`str`, optional):
            prompt (:obj:`str`, optional): Prompt to use. If not provided, uses hwchase17/react.
            chat_history (:obj:`bool`, :obj:`GetSessionHistoryCallable`):
                Whether to use chat history or not. If `True`, uses `InMemoryChatMessageHistory`.
                If a callable is passed, it is used to get the chat history by session_id.
                See Langchain `RunnableWithMessageHistory` get_session_history param for more details.
            output_handler (BaseTool, optional): Tool to use for returning agent's output.
            handle_parsing_errors (:obj:`bool`, optional): Whether to handle parsing errors or not.
            handle_tool_errors (:obj:`bool`, optional): Whether to handle tool errors or not.
                If True, `handle_tool_error` and `handle_validation_error` in all tools are set to True.
            llm (:obj:`BaseLanguageModel`, optional):
            verbose (:obj:`bool`, optional):
        """
        if prompt is None:
            if output_handler is not None:
                raise Exception(OUTPUT_HANDLER_WITH_DEFAULT_PROMPT_MESSAGE)
            prompt = hub.pull("hwchase17/react")

        if llm is None:
            llm = init_llm(llm_framework=LLMFramework.LANGCHAIN)

        if not tools:
            raise ValueError("You must provide at least one tool to the ReActMotleyAgent")

        def agent_factory(
            tools: dict[str, MotleyTool], output_handler: Optional[MotleyTool] = None
        ) -> AgentExecutor:
            langchain_tools = [t.to_langchain_tool() for t in tools.values()]
            if output_handler:
                langchain_tools.append(output_handler.to_langchain_tool())

            if handle_tool_errors:
                for tool in langchain_tools:
                    tool.handle_tool_error = True
                    tool.handle_validation_error = True

            agent = create_react_agent(llm=llm, tools=langchain_tools, prompt=prompt)
            agent_executor = AgentExecutor(
                agent=agent,
                tools=langchain_tools,
                handle_parsing_errors=handle_parsing_errors,
                verbose=verbose,
            )
            return agent_executor

        super().__init__(
            description=description,
            name=name,
            agent_factory=agent_factory,
            tools=tools,
            output_handler=output_handler,
            chat_history=chat_history,
            verbose=verbose,
        )



from .tool import MotleyTool

from .autogen_chat_tool import AutoGenChatTool
from .image.dall_e import DallEImageGeneratorTool
from .llm_tool import LLMTool
from .mermaid_evaluator_tool import MermaidEvaluatorTool
from .python_repl import PythonREPLTool
from .code.postgresql_linter import PostgreSQLLinterTool
from .code.python_linter import PythonLinterTool



""" Module description """
from abc import ABC, abstractmethod
from typing import Optional, Type
from motleycrew.storage import MotleyGraphNode, MotleyGraphNodeType


class MotleyGraphStore(ABC):
    @abstractmethod
    def check_node_exists_by_class_and_id(
        self, node_class: Type[MotleyGraphNode], node_id: int
    ) -> bool:
        """ Check if a node of given class with given id is present in the database.

        Args:
            node_class (Type[MotleyGraphNode]):
            node_id (int):

        Returns:
            bool:
        """
        pass

    @abstractmethod
    def check_node_exists(self, node: MotleyGraphNode) -> bool:
        """ Check if the given node is present in the database.

        Args:
            node (MotleyGraphNode):

        Returns:
            bool:
        """
        pass

    @abstractmethod
    def check_relation_exists(
        self, from_node: MotleyGraphNode, to_node: MotleyGraphNode, label: Optional[str]
    ) -> bool:
        """ Check if a relation exists between two nodes with given label.

        Args:
            from_node (MotleyGraphNode):
            to_node (MotleyGraphNode):
            label (:obj:`str`, None):

        Returns:
            bool:
        """
        pass

    @abstractmethod
    def get_node_by_class_and_id(
        self, node_class: Type[MotleyGraphNodeType], node_id: int
    ) -> Optional[MotleyGraphNodeType]:
        """ Retrieve the node of given class with given id if it is present in the database.
        Otherwise, return None.

        Args:
            node_class (Type[MotleyGraphNodeType]):
            node_id (int):

        Returns:
            :obj:`MotleyGraphNodeType`, None:
        """
        pass

    @abstractmethod
    def insert_node(self, node: MotleyGraphNode):
        """ Insert a new node, populate its id and freeze it.
        If node table or some columns do not exist, this method also creates them.

        Args:
            node (MotleyGraphNode):

        Returns:

        """
        pass

    @abstractmethod
    def create_relation(
        self, from_node: MotleyGraphNode, to_node: MotleyGraphNode, label: str
    ) -> None:
        """ Create a relation with given label between existing nodes.
        If relation table does not exist, this method also creates them.

        Args:
            from_node (MotleyGraphNode):
            to_node (MotleyGraphNode):
            label (str):

        Returns:

        """
        pass

    @abstractmethod
    def upsert_triplet(self, from_node: MotleyGraphNode, to_node: MotleyGraphNode, label: str):
        """ Create a relation with given label between nodes, if such does not already exist.
        If the nodes do not already exist, create them too.
        This method also creates and/or updates all necessary tables.

        Args:
            from_node (MotleyGraphNode):
            to_node (MotleyGraphNode):
            label (str):

        Returns:

        """
        pass

    @abstractmethod
    def delete_node(self, node: MotleyGraphNode) -> None:
        """ Delete a given node and its relations.

        Args:
            node (MotleyGraphNode):

        Returns:

        """
        pass

    @abstractmethod
    def update_property(self, node: MotleyGraphNode, property_name: str):
        """ Update a graph node's property with the corresponding value from the node object.

        Args:
            node (MotleyGraphNode):
            property_name (str):

        Returns:

        """
        pass

    def ensure_node_table(self, node_class: Type[MotleyGraphNode]) -> str:  # for Kuzu
        """ Create a table for storing nodes of that class if such does not already exist.
        If it does exist, create all missing columns.

        Args:
            node_class (Type[MotleyGraphNode]):
        Returns:
            str: Table name
        """
        pass

    def ensure_relation_table(
        self, from_class: Type[MotleyGraphNode], to_class: Type[MotleyGraphNode], label: str
    ):  # for Kuzu
        """ Create a table for storing relations from from_node-like nodes to to_node-like nodes,
        if such does not already exist.

        Args:
            from_class (Type[MotleyGraphNode]):
            to_class (Type[MotleyGraphNode]):
            label (str):

        Returns:

        """
        pass

    def run_cypher_query(
        self,
        query: str,
        parameters: Optional[dict] = None,
        container: Optional[Type[MotleyGraphNodeType]] = None,
    ) -> list[list | MotleyGraphNodeType]:
        """ Run a Cypher query and return the results.
        If container class is provided, deserialize the results into objects of that class.

        Args:
            query (:obj:`dict`, None):
            parameters (:obj:`dict`, optional):
            container (:obj:`Type[MotleyGraphNodeType]`, optional):

        Returns:

        """
        pass



from motleycrew.common.utils import ensure_module_is_installed

try:
    from aider.coders import Coder
    from aider.models import Model
except ImportError:
    Coder = None
    Model = None

from langchain.tools import Tool
from langchain_core.pydantic_v1 import BaseModel, Field

from motleycrew.common import Defaults
from motleycrew.tools import MotleyTool


class AiderTool(MotleyTool):

    def __init__(self, model: str = None, **kwargs):
        """Tool for code generation using Aider.

        Args:
            model (str): model name
            **kwargs:
        """
        ensure_module_is_installed("aider")

        model = model or Defaults.DEFAULT_LLM_NAME
        llm_model = Model(model=model)
        coder = Coder.create(main_model=llm_model, **kwargs)

        langchain_tool = create_aider_tool(coder)
        super(AiderTool, self).__init__(langchain_tool)


class AiderToolInput(BaseModel):
    """Input for the Aider tool.

    Attributes:
        with_message (str):
    """

    with_message: str = Field(description="instructions for code generation")


def create_aider_tool(coder: Coder):
    """Create langchain tool from Aider Coder.run() method

    Returns:
        Tool:
    """
    return Tool.from_function(
        func=coder.run,
        name="aider tool",
        description="Tool for code generation that has access to the provided repository. "
        "Ask it to make changes in the code: fix bugs, add features, write tests etc. "
        "It doesn't run the code by itself.",
        args_schema=AiderToolInput,
    )



class LLMFamily:
    """ Description

    Attributes:
        OPENAI (str):

    """
    OPENAI = "openai"
    ANTHROPIC = "anthropic"


class LLMFramework:
    """ Description

    Attributes:
        LANGCHAIN (str):
        LLAMA_INDEX (str):

    """
    LANGCHAIN = "langchain"
    LLAMA_INDEX = "llama_index"


class GraphStoreType:
    """ Description

    Attributes:
        KUZU (str):

    """
    KUZU = "kuzu"


class TaskUnitStatus:
    """Description

    Attributes:
        PENDING (str):
        RUNNING (str):
        DONE (str):
    """
    PENDING = "pending"
    RUNNING = "running"
    DONE = "done"


class LunaryRunType:
    """ Description

    Attributes:
        LLM (str):
        AGENT (str):
        TOOL (str):
        CHAIN (str):
        EMBED (str):

    """
    LLM = "llm"
    AGENT = "agent"
    TOOL = "tool"
    CHAIN = "chain"
    EMBED = "embed"


class LunaryEventName:
    """ Description

    Attributes:
        START (str):
        END (str):
        UPDATE (str):
        ERROR (str):

    """
    START = "start"
    END = "end"
    UPDATE = "update"
    ERROR = "error"


class AsyncBackend:
    """ Backends for parallel launch

    Attributes:
        ASYNCIO (str): Asynchronous startup using asyncio
        THREADING (str): Running using threads
        NONE (str): Synchronous startup

    """
    ASYNCIO = "asyncio"
    THREADING = "threading"
    NONE = "none"



from .enums import LLMFamily
from .enums import LLMFramework
from .enums import GraphStoreType
from .enums import TaskUnitStatus
from .enums import LunaryRunType
from .enums import LunaryEventName
from .enums import AsyncBackend

from .defaults import Defaults

from .types import MotleySupportedTool
from .types import MotleyAgentFactory

from .logging import logger, configure_logging



""" Module description"""
import sys
from typing import Optional, Sequence
import hashlib
from urllib.parse import urlparse
from langchain_core.messages import BaseMessage

from motleycrew.common.exceptions import ModuleNotInstalled


def to_str(value: str | BaseMessage | Sequence[str] | Sequence[BaseMessage]) -> str:
    """ Description

    Args:
        value (:obj:`str`, :obj:`BaseMessage`, :obj:`Sequence[str]`, :obj:`Sequence[BaseMessage]`):

    Returns:
        str:
    """
    if isinstance(value, str):
        return value
    elif isinstance(value, BaseMessage):
        return value.content
    else:
        try:
            return "\n".join([to_str(v) for v in value])
        except TypeError:
            raise TypeError(f"Expected str, BaseMessage, or an iterable of them, got {type(value)}")


def is_http_url(url):
    """ Description

    Args:
        url (str):

    Returns:
        bool:
    """
    try:
        parsed_url = urlparse(url)
        return parsed_url.scheme in ["http", "https"]
    except ValueError:
        return False


def generate_hex_hash(data: str, length: Optional[int] = None):
    """ Description

    Args:
        data (str):
        length (:obj:`int`, optional):

    Returns:

    """
    hash_obj = hashlib.sha256()
    hash_obj.update(data.encode("utf-8"))
    hex_hash = hash_obj.hexdigest()

    if length is not None:
        hex_hash = hex_hash[:length]
    return hex_hash


def print_passthrough(x):
    return x


def ensure_module_is_installed(module_name: str, install_command: str = None) -> None:
    """ Checking the installation of the module

    Args:
        module_name (str):
        install_command (:obj:`str`, optional):

    Raises:
        ModuleNotInstalled:
    """
    module_path = sys.modules.get(module_name, None)
    if module_path is None:
        raise ModuleNotInstalled(module_name, install_command)



""" Module description"""
from motleycrew.common import Defaults
from motleycrew.common import LLMFamily, LLMFramework
from motleycrew.common.exceptions import LLMFamilyNotSupported, LLMFrameworkNotSupported
from motleycrew.common.utils import ensure_module_is_installed


def langchain_openai_llm(
    llm_name: str = Defaults.DEFAULT_LLM_NAME,
    llm_temperature: float = Defaults.DEFAULT_LLM_TEMPERATURE,
    **kwargs,
):
    """ Description

    Args:
        llm_name (:obj:`str`, optional):
        llm_temperature (:obj:`float`, optional):
        **kwargs:

    Returns:

    """
    from langchain_openai import ChatOpenAI

    return ChatOpenAI(model=llm_name, temperature=llm_temperature, **kwargs)


def llama_index_openai_llm(
    llm_name: str = Defaults.DEFAULT_LLM_NAME,
    llm_temperature: float = Defaults.DEFAULT_LLM_TEMPERATURE,
    **kwargs,
):
    """ Description

    Args:
        llm_name (:obj:`str`, optional):
        llm_temperature (:obj:`float`, optional):
        **kwargs:

    Returns:

    """
    ensure_module_is_installed("llama_index")
    from llama_index.llms.openai import OpenAI

    return OpenAI(model=llm_name, temperature=llm_temperature, **kwargs)


def langchain_anthropic_llm(
    llm_name: str = Defaults.DEFAULT_LLM_NAME,
    llm_temperature: float = Defaults.DEFAULT_LLM_TEMPERATURE,
    **kwargs,
):
    from langchain_anthropic import ChatAnthropic

    return ChatAnthropic(model=llm_name, temperature=llm_temperature, **kwargs)


def llama_index_anthropic_llm(
    llm_name: str = Defaults.DEFAULT_LLM_NAME,
    llm_temperature: float = Defaults.DEFAULT_LLM_TEMPERATURE,
    **kwargs,
):
    ensure_module_is_installed("llama_index")
    from llama_index.llms.anthropic import Anthropic

    return Anthropic(model=llm_name, temperature=llm_temperature, **kwargs)


Defaults.LLM_MAP = {
    (LLMFramework.LANGCHAIN, LLMFamily.OPENAI): langchain_openai_llm,
    (LLMFramework.LLAMA_INDEX, LLMFamily.OPENAI): llama_index_openai_llm,
    (LLMFramework.LANGCHAIN, LLMFamily.ANTHROPIC): langchain_anthropic_llm,
    (LLMFramework.LLAMA_INDEX, LLMFamily.ANTHROPIC): llama_index_anthropic_llm,
}


def init_llm(
    llm_framework: str,
    llm_family: str = Defaults.DEFAULT_LLM_FAMILY,
    llm_name: str = Defaults.DEFAULT_LLM_NAME,
    llm_temperature: float = Defaults.DEFAULT_LLM_TEMPERATURE,
    **kwargs,
):
    """ Description

    Args:
        llm_framework (str):
        llm_family (:obj:`str`, optional):
        llm_name (:obj:`str`, optional):
        llm_temperature (:obj:`float`, optional):
        **kwargs:

    Raises:
        LLMFamilyNotSupported

    Returns:

    """

    func = Defaults.LLM_MAP.get((llm_framework, llm_family), None)
    if func is not None:
        return func(llm_name=llm_name, llm_temperature=llm_temperature, **kwargs)

    raise LLMFamilyNotSupported(llm_framework=llm_framework, llm_family=llm_family)



""" Module description"""
from typing import Any, Optional

from motleycrew.common import Defaults


class LLMFamilyNotSupported(Exception):
    """ Description

    Args:
        llm_framework (str):
        llm_family (str):
    """
    def __init__(self, llm_framework: str, llm_family: str):
        self.llm_framework = llm_framework
        self.llm_family = llm_family

    def __str__(self) -> str:
        return f"LLM family `{self.llm_family}` is not supported via the framework `{self.llm_framework}`"


class LLMFrameworkNotSupported(Exception):
    def __init__(self, llm_framework: str):
        """ Description

        Args:
            llm_framework (str):
        """
        self.llm_framework = llm_framework

    def __str__(self) -> str:
        return f"LLM framework `{self.llm_framework}` is not supported"


class AgentNotMaterialized(Exception):
    """ Description

   Args:
       agent_name (str):
   """
    def __init__(self, agent_name: str):
        self.agent_name = agent_name

    def __str__(self) -> str:
        return f"Agent `{self.agent_name}` is not yet materialized"


class CannotModifyMaterializedAgent(Exception):
    """ Description

    Args:
        agent_name (str):
    """
    def __init__(self, agent_name: str | None):
        self.agent_name = agent_name

    def __str__(self) -> str:
        return "Cannot modify agent{} as it is already materialized".format(
            f" `{self.agent_name}`" if self.agent_name is not None else ""
        )


class TaskDependencyCycleError(Exception):
    """Raised when a task is set to depend on itself"""


class IntegrationTestException(Exception):
    """ Integration tests exception

    Args:
        test_names (list[str]): list of names of failed integration tests
    """

    def __init__(self, test_names: list[str]):
        self.test_names = test_names

    def __str__(self):
        return "Some integration tests failed: {}".format(self.test_names)


class IpynbIntegrationTestResultNotFound(Exception):
    """ Ipynb integration test not found result file exception

    Args:
        ipynb_path (str): path to running ipynb
        result_path (str): path to execution result file
    """
    def __init__(self, ipynb_path: str, result_path: str):
        self.ipynb_path = ipynb_path
        self.result_path = result_path

    def __str__(self):
        return "File result {} of the ipynb {} execution, not found.".format(self.result_path, self.ipynb_path)


class ModuleNotInstalled(Exception):
    """Module not installed

    Args:
        module_name (str): the name of the non-installed module
        install_command (:obj:`str`, optional): the command to install
    """

    def __init__(self, module_name: str, install_command: str = None):
        self.module_name = module_name
        self.install_command = install_command or Defaults.MODULE_INSTALL_COMMANDS.get(
            module_name, None
        )

    def __str__(self):
        msg = "{} is not installed".format(self.module_name)

        if self.install_command is not None:
            msg = "{}, {}".format(msg, self.install_command)

        return "{}.".format(msg)


class InvalidToolInput(Exception):
    """Raised when the tool input is invalid"""

    def __init__(self, tool: Any, input: Any, message: Optional[str] = None):
        self.tool = tool
        self.input = input
        self.message = message

    def __str__(self):
        msg = "Invalid input `{}` for tool `{}`".format(self.input, self.tool_name)
        if self.message:
            msg = "{}: {}".format(msg, self.message)
        return msg


class InvalidOutput(Exception):
    """Raised in output handlers when an agent's output is not accepted"""
    pass



""" Module description """

from typing import Any, Optional, Sequence

try:
    from llama_index.core.agent import AgentRunner
except ImportError:
    AgentRunner = None

from langchain_core.runnables import RunnableConfig

from motleycrew.agents.parent import MotleyAgentParent
from motleycrew.common import MotleySupportedTool
from motleycrew.common import MotleyAgentFactory
from motleycrew.common.utils import ensure_module_is_installed


class LlamaIndexMotleyAgent(MotleyAgentParent):
    def __init__(
        self,
        description: str | None = None,
        name: str | None = None,
        agent_factory: MotleyAgentFactory[AgentRunner] | None = None,
        tools: Sequence[MotleySupportedTool] | None = None,
        verbose: bool = False,
    ):
        """Description

        Args:
            description (:obj:`str`, optional):
            name (:obj:`str`, optional):
            agent_factory (:obj:`MotleyAgentFactory`, optional):
            tools (:obj:`Sequence[MotleySupportedTool]`, optional):
            verbose (:obj:`bool`, optional):
        """
        super().__init__(
            description=description,
            name=name,
            agent_factory=agent_factory,
            tools=tools,
            verbose=verbose,
        )

    def invoke(
        self,
        task_dict: dict,
        config: Optional[RunnableConfig] = None,
        **kwargs: Any,
    ) -> Any:
        """Description

        Args:
            task_dict (dict):
            config (:obj:`RunnableConfig`, optional):
            **kwargs:

        Returns:
            Any:
        """
        self.materialize()
        prompt = self.compose_prompt(task_dict, task_dict.get("prompt"))

        output = self.agent.chat(prompt)
        return output.response

    @staticmethod
    def from_agent(
        agent: AgentRunner,
        description: Optional[str] = None,
        tools: Sequence[MotleySupportedTool] | None = None,
        verbose: bool = False,
    ) -> "LlamaIndexMotleyAgent":
        """Description

        Args:
            agent (AgentRunner):
            description (:obj:`str`, optional):
            tools (:obj:`Sequence[MotleySupportedTool]`, optional):
            verbose (:obj:`bool`, optional):

        Returns:
            LlamaIndexMotleyAgent:
        """
        ensure_module_is_installed("llama_index")
        wrapped_agent = LlamaIndexMotleyAgent(description=description, tools=tools, verbose=verbose)
        wrapped_agent._agent = agent
        return wrapped_agent



""" Module description

Attributes:
    IS_SUBQUESTION_PREDICATE (str):
    default_prompt (PromptTemplate):

"""

from typing import Optional
from pathlib import Path
import time

from langchain_core.language_models import BaseLanguageModel
from langchain_core.runnables import (
    RunnablePassthrough,
    RunnableLambda,
)
from langchain_core.tools import Tool
from langchain_core.prompts.base import BasePromptTemplate
from langchain_core.prompts import PromptTemplate

from langchain_core.pydantic_v1 import BaseModel, Field

from motleycrew.tools import MotleyTool
from motleycrew.common import LLMFramework
from motleycrew.common.llms import init_llm
from motleycrew.common.utils import print_passthrough
from motleycrew.storage import MotleyGraphStore
from motleycrew.common import logger


from motleycrew.applications.research_agent.question import Question, QuestionGenerationTaskUnit

IS_SUBQUESTION_PREDICATE = "is_subquestion"

default_prompt = PromptTemplate.from_template(
    """
You are a part of a team. The ultimate goal of your team is to
answer the following Question: '{question}'.\n
Your team has discovered some new text (delimited by ```) that may be relevant to your ultimate goal.
text: \n ``` {context} ``` \n
Your task is to ask new questions that may help your team achieve the ultimate goal.
If you think that the text is relevant to your ultimate goal, then ask new questions.
New questions should be based only on the text and the goal Question and no other previous knowledge.

You can ask up to {num_questions} new questions.
Return the questions each on a new line and ending with a single question mark.
Don't return anything else except these questions.
"""
)

# " The new questions should have no semantic overlap with questions in the following list:\n"
# " {previous_questions}\n"


class QuestionGeneratorTool(MotleyTool):
    """
    Gets a question as input
    Retrieves relevant docs (llama index basic RAG)
    (Retrieves existing questions from graph (to avoid overlap))
    Generates extra questions (research agent prompt)

    Adds questions as children of current q by calling Q insertion tool once
    exits
    """

    def __init__(
        self,
        query_tool: MotleyTool,
        graph: MotleyGraphStore,
        max_questions: int = 3,
        llm: Optional[BaseLanguageModel] = None,
        prompt: str | BasePromptTemplate = None,
    ):
        """Description

        Args:
            query_tool (MotleyTool):
            graph (MotleyGraphStore):
            max_questions (:obj:`int`, optional):
            llm (:obj:`BaseLanguageModel`, optional:
            prompt (:obj:`str`, :obj:`BasePromptTemplate`, optional):
        """
        langchain_tool = create_question_generator_langchain_tool(
            query_tool=query_tool,
            graph=graph,
            max_questions=max_questions,
            llm=llm,
            prompt=prompt,
        )

        super().__init__(langchain_tool)


class QuestionGeneratorToolInput(BaseModel, arbitrary_types_allowed=True):
    """Input for the Question Generator Tool.

    Attributes:
        question (Question):
    """

    question: Question = Field(description="The input question for which to generate subquestions.")


def create_question_generator_langchain_tool(
    query_tool: MotleyTool,
    graph: MotleyGraphStore,
    max_questions: int = 3,
    llm: Optional[BaseLanguageModel] = None,
    prompt: str | BasePromptTemplate = None,
):
    """Description

    Args:
        query_tool (MotleyTool):
        graph (MotleyGraphStore):
        max_questions (:obj:`int`, optional):
        llm (:obj:`BaseLanguageModel`, optional:
        prompt (:obj:`str`, :obj:`BasePromptTemplate`, optional):

    Returns:

    """
    if llm is None:
        llm = init_llm(llm_framework=LLMFramework.LANGCHAIN)

    llm.bind(json_mode=True)

    if prompt is None:
        prompt = default_prompt
    elif isinstance(prompt, str):
        prompt = PromptTemplate.from_template(prompt)

    assert isinstance(prompt, BasePromptTemplate), "Prompt must be a string or a BasePromptTemplate"

    def insert_questions(input_dict) -> None:
        questions_raw = input_dict["subquestions"].content
        questions = [q.strip() for q in questions_raw.split("\n") if len(q.strip()) > 1]
        for q in questions:
            logger.info("Inserting question: %s", q)
            subquestion = graph.insert_node(Question(question=q))
            graph.create_relation(input_dict["question"], subquestion, IS_SUBQUESTION_PREDICATE)
        logger.info("Inserted %s questions", len(questions))

    def set_context(input_dict: dict):
        node = input_dict["question"]
        node.context = input_dict["context"]

    pipeline = (
        RunnableLambda(print_passthrough)
        | RunnablePassthrough().assign(context=query_tool.to_langchain_tool())
        | RunnableLambda(print_passthrough)
        | RunnablePassthrough().assign(
            subquestions=prompt.partial(num_questions=str(max_questions)) | llm
        )
        | RunnableLambda(print_passthrough)
        | {
            "set_context": RunnableLambda(set_context),
            "insert_questions": RunnableLambda(insert_questions),
        }
    )

    return Tool.from_function(
        func=lambda q: pipeline.invoke({"question": q}),
        name="Question Generator Tool",
        description="""Generate a list of questions based on the input question, 
    and insert them into the knowledge graph.""",
        args_schema=QuestionGeneratorToolInput,
    )


if __name__ == "__main__":
    import kuzu
    from llama_index.graph_stores.kuzu import KuzuGraphStore

    here = Path(__file__).parent
    db_path = str(here / "test2")

    db = kuzu.Database(db_path)
    graph_store = KuzuGraphStore(db)

    query_tool = MotleyTool.from_langchain_tool(
        Tool.from_function(
            func=lambda question: [
                "Germany has consisted of many different states over the years",
                "The capital of France has moved in 1815, from Lyons to Paris",
                "France actually has two capitals, one in the north and one in the south",
            ],
            name="Query Tool",
            description="Query the library for relevant information.",
            args_schema=QuestionGeneratorToolInput,
        )
    )

    tool = QuestionGeneratorTool(
        query_tool=query_tool,
        graph=graph_store,
        max_questions=3,
    )

    tool.invoke({"question": "What is the capital of France?"})
    print("Done!")



""" Module description """
import os.path

from langchain_core.pydantic_v1 import BaseModel, Field
from langchain_core.tools import StructuredTool

from llama_index.core.node_parser import SentenceSplitter
from llama_index.embeddings.openai import OpenAIEmbedding

from llama_index.core import (
    VectorStoreIndex,
    SimpleDirectoryReader,
    StorageContext,
    load_index_from_storage,
)

from motleycrew.tools import MotleyTool
from motleycrew.applications.research_agent.question import Question


class SimpleRetrieverTool(MotleyTool):
    def __init__(self, DATA_DIR, PERSIST_DIR, return_strings_only: bool = False):
        """ Description

        Args:
            DATA_DIR (str):
            PERSIST_DIR (str):
            return_strings_only (:obj:`bool`, optional):
        """
        tool = make_retriever_langchain_tool(
            DATA_DIR, PERSIST_DIR, return_strings_only=return_strings_only
        )
        super().__init__(tool)


class RetrieverToolInput(BaseModel, arbitrary_types_allowed=True):
    """Input for the Retriever Tool.

    Attributes:
        question (Question):

    """

    question: Question = Field(
        description="The input question for which to retrieve relevant data."
    )


def make_retriever_langchain_tool(DATA_DIR, PERSIST_DIR, return_strings_only: bool = False):
    """ Description

    Args:
        DATA_DIR (str):
        PERSIST_DIR (str):
        return_strings_only (:obj:`bool`, optional):

    Returns:

    """
    text_embedding_model = "text-embedding-ada-002"
    embeddings = OpenAIEmbedding(model=text_embedding_model)

    if not os.path.exists(PERSIST_DIR):
        # load the documents and create the index
        documents = SimpleDirectoryReader(DATA_DIR).load_data()
        index = VectorStoreIndex.from_documents(
            documents, transformations=[SentenceSplitter(chunk_size=512), embeddings]
        )
        # store it for later
        index.storage_context.persist(persist_dir=PERSIST_DIR)
    else:
        # load the existing index
        storage_context = StorageContext.from_defaults(persist_dir=PERSIST_DIR)
        index = load_index_from_storage(storage_context)

    retriever = index.as_retriever(
        similarity_top_k=10,
        embed_model=embeddings,
    )

    def call_retriever(question: Question) -> list:
        out = retriever.retrieve(question.question)
        if return_strings_only:
            return [node.text for node in out]
        return out

    retriever_tool = StructuredTool.from_function(
        func=call_retriever,
        name="information_retriever_tool",
        description="Useful for running a natural language query against a"
        " knowledge base and retrieving a set of relevant documents.",
        args_schema=RetrieverToolInput,
    )
    return retriever_tool


if __name__ == "__main__":

    # check if storage already exists
    here = os.path.dirname(os.path.abspath(__file__))
    DATA_DIR = os.path.join(here, "mahabharata/text/TinyTales")

    PERSIST_DIR = "../../examples/research_agent/storage"

    retriever_tool = SimpleRetrieverTool(DATA_DIR, PERSIST_DIR)
    response2 = retriever_tool.invoke(
        {"question": Question(question="What are the most interesting facts about Arjuna?")}
    )
    print(response2)



""" Module description"""
from typing import TYPE_CHECKING, Union, Optional, Protocol, TypeVar

if TYPE_CHECKING:
    from langchain.tools import BaseTool

    try:
        from llama_index.core.tools import BaseTool as LlamaIndex__BaseTool
    except ImportError:
        LlamaIndex__BaseTool = "LlamaIndex__BaseTool"

    from motleycrew.tools import MotleyTool
    from motleycrew.agents.abstract_parent import MotleyAgentAbstractParent

else:
    MotleyTool = "MotleyTool"
    BaseTool = "BaseTool"
    LlamaIndex__BaseTool = "LlamaIndex__BaseTool"
    MotleyAgentAbstractParent = "MotleyAgentAbstractParent"


MotleySupportedTool = Union[MotleyTool, BaseTool, LlamaIndex__BaseTool, MotleyAgentAbstractParent]


AgentType = TypeVar("AgentType")


class MotleyAgentFactory(Protocol[AgentType]):
    """
    Type protocol for an agent factory.
    It is a function that accepts tools as an argument
    and returns an agent instance of an appropriate class.
    """

    def __call__(
        self,
        tools: dict[str, MotleyTool],
        output_handler: Optional[MotleyTool] = None,
    ) -> AgentType: ...



""" Module description"""
from abc import ABC, abstractmethod
from typing import Optional, Any, TYPE_CHECKING

from langchain_core.runnables import RunnableConfig

if TYPE_CHECKING:
    from motleycrew.tools import MotleyTool


class MotleyAgentAbstractParent(ABC):
    @abstractmethod
    def invoke(
        self,
        task_dict: dict,
        config: Optional[RunnableConfig] = None,
        **kwargs: Any,
    ) -> Any:
        """ Description

            Args:
                task_dict (dict):
                config (:obj:`RunnableConfig`, optional):
                **kwargs:

            Returns:
                Any:
        """
        pass

    @abstractmethod
    def as_tool(self) -> "MotleyTool":
        """ Description

        Returns:
            MotleyTool

        """
        pass



from typing import Sequence, List, Optional

from langchain_core.messages import BaseMessage, HumanMessage, ToolMessage
from langchain_core.language_models import BaseChatModel
from langchain_core.prompts.chat import ChatPromptTemplate
from langchain_core.runnables import Runnable, RunnablePassthrough, RunnableLambda
from langchain_core.runnables.history import GetSessionHistoryCallable
from langchain_core.tools import BaseTool
from langchain_core.agents import AgentFinish, AgentActionMessageLog
from langchain_core.prompts import MessagesPlaceholder
from langchain_core.messages.base import merge_content

from langchain.agents import AgentExecutor
from langchain.agents.format_scratchpad.tools import format_to_tool_messages
from langchain.agents.output_parsers.tools import ToolsAgentOutputParser

from langchain.tools.render import render_text_description

from motleycrew.agents.langchain import LangchainMotleyAgent
from motleycrew.tools import MotleyTool
from motleycrew.common import MotleySupportedTool
from motleycrew.common import LLMFramework
from motleycrew.common.llms import init_llm
from motleycrew.common.utils import print_passthrough


default_think_prompt = ChatPromptTemplate.from_messages(
    [
        (
            "system",
            """
Answer the following questions as best you can; think carefully, one step at a time, 
and outline the next step towards answering the question.

You will later have access to the following tools, but you can't use them yet:
{tools}

Your reply must begin with "Thought:" and then describe what the next step should be, 
given the information so far - either to use one of the tools from the above list 
(and details on how to use it), or give the final answer like described below.
Do NOT return a tool call, just a single thought about what it should be.
If the information so far is not sufficient to answer the question precisely and completely
(rather than sloppily and approximately), don't hesitate to
request to use tools again, until sufficient information is gathered. Don't stop this until
you are certain that you have enough information to answer the question.

If you have sufficient information to answer the question, your reply must look like
```
Thought: I now know the final answer
Final Answer: [the final answer to the original input question]
```
but without the backticks.

Begin!

""",
        ),
        MessagesPlaceholder(variable_name="chat_history", optional=True),
        ("user", "{input}"),
        MessagesPlaceholder(variable_name="agent_scratchpad"),
    ]
)


default_act_prompt = ChatPromptTemplate.from_messages(
    [
        (
            "system",
            """
Your objective is to contribute to answering the below question, by using the tools at your disposal.

You have access to the following tools:
{tools}

Your response MUST be only a tool call to one of these unless the last input
message contains the words "Final Answer"; {output_instruction}
{output_handler}

""",
        ),
        MessagesPlaceholder(variable_name="chat_history", optional=True),
        ("user", "{input}"),
        MessagesPlaceholder(variable_name="agent_scratchpad"),
    ]
)
default_act_prompt_without_output_handler = default_act_prompt.partial(
    output_instruction="if it does, just repeat it."
)
default_act_prompt_with_output_handler = default_act_prompt.partial(
    output_instruction="if it does, you must use the output handler:"
)


def check_variables(prompt: ChatPromptTemplate):
    """
    Args:
        prompt (ChatPromptTemplate):

    Returns:

    """
    missing_vars = {"agent_scratchpad"}.difference(prompt.input_variables)
    if missing_vars:
        raise ValueError(f"Prompt missing required variables: {missing_vars}")


def add_thought_to_background(x: dict):
    """Description

    Args:
        x (dict):

    Returns:

    """
    out = x["background"]
    out["agent_scratchpad"] += [x["thought"]]
    return out


def cast_thought_to_human_message(thought: BaseMessage):
    content = thought.content
    if content:
        # HACK: this cast to a HumanMessage is needed for Anthropic models to work
        # because they will treat an AIMessage in input as their output prefill.
        # Also, we remove tool_use messages from the content in case a model calls a tool.
        if isinstance(content, list):
            content = [chunk for chunk in content if chunk.get("type") != "tool_use"]
    return HumanMessage(content=content)


def add_messages_to_action(
    actions: List[AgentActionMessageLog] | AgentFinish, messages: List[BaseMessage]
) -> List[AgentActionMessageLog] | AgentFinish:
    """
    Args:
        actions (:obj:`List[AgentActionMessageLog]`, :obj:`AgentFinish`):
        messages (List[BaseMessage]):

    Returns:
        List[AgentActionMessageLog] | AgentFinish:
    """
    if not isinstance(actions, AgentFinish):
        for action in actions:
            action.message_log = messages + list(action.message_log)
    return actions


def merge_consecutive_messages(messages: Sequence[BaseMessage]) -> List[BaseMessage]:
    """Tries to merge consecutive messages of the same type in the provided list.
    This mainly serves as a workaround for Anthropic models that don't accept
    multiple AIMessages in a row.

    Args:
        messages (Sequence[BaseMessage]): The list of messages to process.

    Returns:
        List[BaseMessage]: The list of messages with consecutive messages of the same type merged.
    """
    merged_messages = []
    for message in messages:
        if (
            not merged_messages
            or type(merged_messages[-1]) != type(message)
            or isinstance(message, ToolMessage)
        ):
            merged_messages.append(message)
        else:
            merged_messages[-1].content = merge_content(
                merged_messages[-1].content, message.content
            )
    return merged_messages


def create_tool_calling_react_agent(
    llm: BaseChatModel,
    tools: Sequence[BaseTool],
    think_prompt: ChatPromptTemplate | None = None,
    act_prompt: ChatPromptTemplate | None = None,
    output_handler: BaseTool | None = None,
) -> Runnable:
    """Create a ReAct-style agent that supports tool calling.

    Args:
        llm (BaseChatModel): LLM to use as the agent.
        tools (Sequence[BaseTool]): Tools this agent has access to.
        think_prompt (ChatPromptTemplate, optional): The thinking step prompt to use.
            See Prompt section below for more on the expected input variables.
        act_prompt (ChatPromptTemplate, optional): The acting step prompt to use.
            See Prompt section below for more on the expected input variables.
        output_handler (BaseTool, optional): Tool to use for returning agent's output.

    Returns:
        A Runnable sequence representing an agent. It takes as input all the same input
        variables as the prompt passed in does. It returns as output either an
        AgentAction or AgentFinish.

    Prompt:
        This agent uses two prompts, one for thinking and one for acting. The prompts
        must have `agent_scratchpad` and `chat_history` ``MessagesPlaceholder``s.
        If a prompt is not passed in, the default one is used.

    """
    if think_prompt is None:
        think_prompt = default_think_prompt

    if act_prompt is None:
        if output_handler:
            act_prompt = default_act_prompt_with_output_handler
        else:
            act_prompt = default_act_prompt_without_output_handler

    think_prompt = think_prompt.partial(tools=render_text_description(list(tools)))
    act_prompt = act_prompt.partial(
        tools=render_text_description(list(tools)),
        output_handler=render_text_description([output_handler]) if output_handler else "",
    )

    check_variables(think_prompt)
    check_variables(act_prompt)

    tools_for_llm = list(tools)
    if output_handler:
        tools_for_llm.append(output_handler)
    llm_with_tools = llm.bind_tools(tools=tools_for_llm)

    think_chain = think_prompt | llm_with_tools | RunnableLambda(cast_thought_to_human_message)
    act_chain = (
        RunnableLambda(add_thought_to_background)
        | act_prompt
        | llm_with_tools
        | ToolsAgentOutputParser()
    )

    agent = (
        RunnablePassthrough.assign(
            agent_scratchpad=lambda x: merge_consecutive_messages(
                format_to_tool_messages(x["intermediate_steps"])
            )
        )
        | {"thought": think_chain, "background": RunnablePassthrough()}
        | RunnableLambda(print_passthrough)
        | {"action": act_chain, "thought": RunnableLambda(lambda x: x["thought"])}
        | RunnableLambda(lambda x: add_messages_to_action(x["action"], [x["thought"]]))
    )
    return agent


class ReActToolCallingAgent(LangchainMotleyAgent):
    def __init__(
        self,
        tools: Sequence[MotleySupportedTool],
        description: str | None = None,
        name: str | None = None,
        think_prompt: ChatPromptTemplate | None = None,
        act_prompt: ChatPromptTemplate | None = None,
        chat_history: bool | GetSessionHistoryCallable = True,
        output_handler: MotleySupportedTool | None = None,
        handle_parsing_errors: bool = True,
        handle_tool_errors: bool = True,
        llm: BaseChatModel | None = None,
        verbose: bool = False,
    ):
        """Universal ReAct-style agent that supports tool calling.

        Args:
            tools (Sequence[MotleySupportedTool]):
            description (:obj:`str`, optional):
            name (:obj:`str`, optional):
            think_prompt (ChatPromptTemplate, optional): The thinking step prompt to use.
                See Prompt section below for more on the expected input variables.
            act_prompt (ChatPromptTemplate, optional): The acting step prompt to use.
                See Prompt section below for more on the expected input variables.
            chat_history (:obj:`bool`, :obj:`GetSessionHistoryCallable`):
                Whether to use chat history or not. If `True`, uses `InMemoryChatMessageHistory`.
                If a callable is passed, it is used to get the chat history by session_id.
                See Langchain `RunnableWithMessageHistory` get_session_history param for more details.
            output_handler (BaseTool, optional): Tool to use for returning agent's output.
            handle_parsing_errors (:obj:`bool`, optional): Whether to handle parsing errors or not.
            handle_tool_errors (:obj:`bool`, optional): Whether to handle tool errors or not.
                If True, `handle_tool_error` and `handle_validation_error` in all tools are set to True.
            llm (:obj:`BaseLanguageModel`, optional):
            verbose (:obj:`bool`, optional):
        """
        if llm is None:
            llm = init_llm(llm_framework=LLMFramework.LANGCHAIN)

        if not tools:
            raise ValueError("You must provide at least one tool to the ReActToolCallingAgent")

        def agent_factory(
            tools: dict[str, MotleyTool], output_handler: Optional[MotleyTool] = None
        ) -> AgentExecutor:
            tools_for_langchain = [t.to_langchain_tool() for t in tools.values()]
            if output_handler:
                output_handler_for_langchain = output_handler.to_langchain_tool()
            else:
                output_handler_for_langchain = None

            agent = create_tool_calling_react_agent(
                llm=llm,
                tools=tools_for_langchain,
                think_prompt=think_prompt,
                act_prompt=act_prompt,
                output_handler=output_handler_for_langchain,
            )

            if output_handler_for_langchain:
                tools_for_langchain.append(output_handler_for_langchain)

            if handle_tool_errors:
                for tool in tools_for_langchain:
                    tool.handle_tool_error = True
                    tool.handle_validation_error = True

            agent_executor = AgentExecutor(
                agent=agent,
                tools=tools_for_langchain,
                handle_parsing_errors=handle_parsing_errors,
                verbose=verbose,
            )
            return agent_executor

        super().__init__(
            description=description,
            name=name,
            agent_factory=agent_factory,
            tools=tools,
            output_handler=output_handler,
            chat_history=chat_history,
            verbose=verbose,
        )
